#ifndef lint
static char *RCSid = "$Header: /sequent2/empire/EMP/lib/common/RCS/damage.c,v 1.3 89/09/17 22:31:18 mr-frog Exp $";
#endif

/*
 * damage.c
 *
 * damage a sector, ship, or plane.
 *
 * Should probably do this in some regular fashion, but
 * we don't yet have a list of fields we can iterate through
 * and find the "damageable" ones.
 *
 * Dave Pare, 1989
 */

#include <ctype.h>
#include <math.h>
#include "misc.h"
#include "var.h"
#include "spy.h"
#include "sect.h"
#include "ship.h"
#include "xy.h"
#include "nat.h"
#include "plane.h"
#include "nuke.h"
#include "file.h"

extern double bunker_strength;

sectdamage(sp,dam)
register struct sctstr *sp;
int dam;
{
    sectdam(sp,dam,0);
}

sectneutrondamage(sp,dam)
register struct sctstr *sp;
int dam;
{
    sectdam(sp,dam,1);
}

sectdam(sp,dam,neutron)
register struct sctstr *sp;
int dam,neutron;
{
    /*
    **  4/19/91 bf@hp - modified to prevent damage to mil in bunkers
    **	(but note that the bunker itself will still sustain damage).
    */
    int mil,milabove,milbelow;
    if (dam <= 0) return;
    if (dam > 100) dam = 100;
    sp->sct_effic = damage((int)sp->sct_effic, dam);
    sp->sct_mobil = damage((int)sp->sct_mobil, dam);
#ifdef BUNKERS
    mil = getvar(V_MILIT, (char *)sp, EF_SECTOR);
    milabove = mil - sp->sct_harden;
    if (milabove<0) milabove = 0;
    milbelow = mil - milabove;
    putvar(V_MILIT, milabove, (char *)sp, EF_SECTOR);
#endif
    sp->sct_nv = neutron 
	? vl_neutrondamage(dam, sp->sct_vtype, sp->sct_vamt,(int)sp->sct_nv)
        : vl_damage(dam, sp->sct_vtype, sp->sct_vamt,(int)sp->sct_nv);
#ifdef BUNKERS
    milabove = getvar(V_MILIT, (char *)sp, EF_SECTOR);
    putvar(V_MILIT, milabove+milbelow, (char *)sp, EF_SECTOR);
    dam = (int)((double)dam/bunker_strength);
    sp->sct_harden = damage((int)sp->sct_harden, dam);
#endif
    if (sp->sct_effic < 20 && sp->sct_type == SCT_BHEAD)
	bridgefall(sp);
}

shipdamage(sp,dam)
struct shpstr *sp;
int dam;
{	
    shipdam(sp,dam,0);
}

shipneutrondamage(sp,dam)
struct shpstr *sp;
int dam;
{
    shipdam(sp,dam,1);
}

shipdam(sp,dam,neutron)
struct shpstr *sp;
int dam,neutron;
{
    if (dam <= 0) return;
    if (dam > 100) dam = 100;
    sp->shp_effic = damage((int)sp->shp_effic, dam);
    if (sp->shp_mobil > 0)
	sp->shp_mobil = damage((int)sp->shp_mobil, dam);
    sp->shp_nv = neutron
	? vl_neutrondamage(dam, sp->shp_vtype, sp->shp_vamt,(int) sp->shp_nv)
	: vl_damage(dam, sp->shp_vtype, sp->shp_vamt,(int) sp->shp_nv);
}

spydamage(sp,dam)
struct spystr *sp;
int dam;
{
    spydam(sp,dam,0);
}

spyneutrondamage(sp,dam)
struct spystr *sp;
int dam;
{
    spydam(sp,dam,1);
}

spydam(sp,dam,neutron)
struct spystr *sp;
int dam,neutron;
{
    if (dam <= 0) return;
    if (dam > 100) dam = 100;
    if (neutron)
	sp->spy_effic = 0;
    else {
	sp->spy_effic = damage((int)sp->spy_effic, dam);
	if (sp->spy_mobil > 0)
	    sp->spy_mobil = damage((int)sp->spy_mobil, dam);
    }
}

planedamage(pp, dam)
struct	plnstr *pp;
int	dam;
{
    if (dam <= 0) return;
    if (dam > 100) dam = 100;
    pp->pln_effic = damage((int)pp->pln_effic, dam);
    if (pp->pln_mobil > 0)
	pp->pln_mobil = damage((int)pp->pln_mobil, dam);
}

/*
**  nukedamage() actually just calculates damage
**  rather than inflicting it.
*/

#ifdef OLDNUKEDAMAGE
int nukedamage(ncp, range, airburst)
struct	nchrstr *ncp;
int	range,airburst;
{
    int	dam;
    int	rad;
    
    rad = ncp->n_blast;
    if (airburst) rad = (int) (rad * 1.5);
    if (rad < range) return 0;
    if (airburst) {
	/* larger area, less center damage */
	dam = (int) ((ncp->n_dam * 0.75) - (range * 20));
    } else {
	/* smaller area, more center damage */
	dam = (int) (ncp->n_dam / (range + 1.0));
    }
    if (dam < 5) dam = 0;
    return dam;
}
#else
int nukedamage(ncp, range, airburst)
struct nchrstr *ncp;
int range,airburst;
{
    int dam;
    double r,radius,spread,damage;
    damage = (double)ncp->n_dam;
    spread = (double)ncp->n_blast;
    radius = (double)range;
    if (airburst) {
	r = exp(radius/spread);
	dam = (int)((damage*0.75)/(r*r) + 0.5);
    } else {
	r = exp(radius*radius/spread);
	dam = (int)(damage/r + 0.5);
    }
    if (dam < 5) dam = 0;
    return dam;
}
#endif

int damage(amt, pct)
register int amt;
int pct;
{
    register int tmp;
    register int lost;
    
    if (amt <= 0) return 0;
    tmp = amt * pct;
    lost = tmp / 100;
    if ((random() % 100) < (tmp % 100)) lost++;
    return amt - lost;
}

/*
**  cannotleave - Check ability of a vehicle to leave a sector,
**  adjusting mobility and petrol costs, if necessary.
**
**  This probably doesn't belong in this module, but it *is* a 
**  form of damage, and this routine needs to be in a common place
**  since it gets used by both the player and update code.
*/

int cannotleave(shpp,mobcost,petcost,notifydriver,notifysector)
struct shpstr *shpp;
int notifydriver,notifysector;
double *mobcost;
int *petcost;
{
    coord x,y;
    int pcost,mil,civ,retcode;
    double mcost;
    struct sctstr sect;
    struct mchrstr *mcp;

    if (!getsect(shpp->shp_x,shpp->shp_y,&sect)) return 0;

    if (sect.sct_own == shpp->shp_own ||
	getrel(getnatp(cnum),sect.sct_own) == ALLIED) return 0;
	    
    mil = getvar(V_MILIT,(char *)&sect,EF_SECTOR);
    civ = getvar(V_CIVIL,(char *)&sect,EF_SECTOR);
    mcp = &mchr[shpp->shp_type];
    x = shpp->shp_x;
    y = shpp->shp_y;

    /*
    **  TRAINS
    **  Trains are hard NOT to see, so we just keep tabs on them.
    */

    if (istrain(shpp)) {
	retcode = 0;
	if (mil>50 && chance(mil/200)) {
	    retcode = 1;
	    mcost = (*mobcost)*4;
	    pcost = (*petcost)*2;
	    switch (notifydriver) {
	    case 1:
		wu(0,shpp->shp_own,
		   fmt("%s #%d stopped by military in %s\n",
		       mcp->m_name,shpp->shp_uid,xyas(x,y,shpp->shp_own)));
		break;
	    case 2:
		pr(fmt("%s #%d stopped by local military\n",
		       mcp->m_name,shpp->shp_uid));
		break;
	    }
	    *mobcost = mcost;
	    *petcost = pcost;
	}
	if (notifysector && (chance((civ+mil*2)/50) || retcode>0)) {
	    wu(0,sect.sct_own,
	       fmt("%s #%d seen in %s\n",
		   mcp->m_name,shpp->shp_uid,xyas(x,y,sect.sct_own)));
	}
	return retcode;
    }

    /*
    **  TANK TRYING TO LEAVE AN ENEMY CONTROLLED SECTOR
    **  Inform the sector's owner that there's a tank in his midst.
    **  If there are more than 20 military there, tell him it's not possible 
    **  and don't bother trying.  Otherwise, give it a shot.  There is a 
    **  small chance, inversely proportional to the number of military in
    **  the sector, that the tank will get out scott free; more than likely, 
    **  though, the guy will just burn 2-4x normal mobility evading ground
    **  troops.
    */

    if (istank(shpp)) {
	if (notifysector && chance(mil/60.0))
	    wu(0,sect.sct_own,
	       fmt("%s #%d spotted in %s!\n",
		   mcp->m_name,shpp->shp_uid,xyas(x,y,sect.sct_own)));
	
	if (mil >= 20) {
	    switch (notifydriver) {
	    case 1:
		wu(0,shpp->shp_own,
		   fmt("%s #%d blocked in a military control zone at %s\n",
		       mcp->m_name,shpp->shp_uid,xyas(x,y,shpp->shp_own)));
		break;
	    case 2:
		pr(fmt("%s #%d blocked in a military control zone\n",
		       mcp->m_name,shpp->shp_uid));
		break;
	    }
	    return 1;
	}
	
	if (roll(mil+1)>1) {
	    /* charge 2-4x normal mobility cost */
	    mcost = (*mobcost)*(double)(roll(17)+15)/8.0;
	    /* burn 2-3x normal petrol cost */
	    pcost = (*petcost)*(int)(roll(2)+1);
	    switch (notifydriver) {
	    case 1:
		wu(0,shpp->shp_own,
		   fmt("%s #%d evading ground troops in %s\n",
		       mcp->m_name,shpp->shp_uid,
		       xyas(x,y,shpp->shp_own)));
		break;
	    case 2:
		pr(fmt("%s #%d used additional %d pet %.1f mob evading ground troops\n",
		       mcp->m_name,
		       shpp->shp_uid,
		       pcost-(*petcost),
		       mcost-(*mobcost)));
		break;
	    }
	    *mobcost = mcost;
	    *petcost = pcost;
	}
    }
    return 0;
}    
