/* _________________________________________________________________
**|                                                                 |
**| Names/definitions support for Empire                            |
**|                                                                 |
**| All of these routines except for opennames() and closenames()   |
**| return TRUE for success and FALSE if they fail.                 |
**|                                                                 |
**| 7/24/90 bf@hp - Initial version                                 |
**|_________________________________________________________________|
*/

#include <stdio.h>
#include <fcntl.h>

#include "gamesdef.h"
#include "misc.h"

#define MAXNAMELEN	24
#define MAXDEFNLEN	40
#define BUFLEN		(MAXNAMELEN+MAXDEFNLEN)

int namesf;
char namesfil[] = EMPPATH(data/names);
char linebuf[BUFLEN];

/* ________________________________________________________________
**|                                                                |
**| opennames(cnum) - open the names file for country 'cnum'       |
**|________________________________________________________________|
*/

opennames(cnum)
int cnum;
{
    char name[80];
    sprintf(name,"%s%02d",namesfil,cnum);
    namesf = open(name, O_CREAT | O_RDWR, 0600);
}

/* ________________________________________________________________
**|                                                                |
**| closenames() - close the currently open names file             |
**|________________________________________________________________|
*/

closenames()
{
    if (namesf>=0) close(namesf);
    namesf = -1;
}

/* ________________________________________________________________
**|                                                                |
**| addname(s,d) - addname 's' and definition 'd' to the list      |
**|________________________________________________________________|
*/

addname(s,d)
char *s,*d;
{
    int i;
    long j;
    delname(s);
    if (namesf>=0) {
	lseek(namesf,(j=0L),0);
	while (1) {
	    i = read(namesf,linebuf,BUFLEN);
	    if (i==BUFLEN && *linebuf) j += (long)BUFLEN;
	    else break;
	}
	if (i==BUFLEN) lseek(namesf,j,0);
	strncpy(linebuf,s,MAXNAMELEN);
	linebuf[MAXNAMELEN-1] = '\0';
	strncpy(linebuf+MAXNAMELEN,d,MAXDEFNLEN);
	linebuf[BUFLEN-1] = '\0';
	write(namesf,linebuf,BUFLEN);
	return(1);
    }
    return(0);
}

/* ________________________________________________________________
**|                                                                |
**| delname(s) - delete name 's' from the list                     |
**|________________________________________________________________|
*/

delname(s)
char *s;
{
    int i;
    long j;
    if (namesf>=0) {
	lseek(namesf,(j=0L),0);
	while (1) {
	    i = read(namesf,linebuf,BUFLEN);
	    if (i==BUFLEN && strcmp(linebuf,s)) j += (long)BUFLEN;
	    else break;
	}
	if (i==BUFLEN) {
	    *linebuf = '\0';
	    lseek(namesf,j,0);
	    write(namesf,linebuf,1);
	    return(1);
	}
    }
    return(0);
}

/* ____________________________________________________________________
**|                                                                    |
**| listnames(p) - list all names/definitions that match pattern 'p'   |
**|____________________________________________________________________|
*/

listnames(p)
char *p;
{
    int i;
    long j;
    if (namesf>=0) {
	lseek(namesf,(j=0L),0);
	do {
	    i = read(namesf,linebuf,BUFLEN);
	    if (i==BUFLEN && patmat(linebuf,p,0))
		pr(fmt("%24s %40s\n",&linebuf[0],&linebuf[MAXNAMELEN]));
	} while (i==BUFLEN);
	return(1);
    }
    return(0);
}

/* ________________________________________________________________
**|                                                                |
**| findname(s,d) - return definition of name 's' in string 'd'    |
**|________________________________________________________________|
*/

findname(s,d)
char *s,*d;
{
    int i;
    long j;
    *d = '\0';
    if (namesf>=0) {
	lseek(namesf,(j=0L),0);
	while (1) {
	    i = read(namesf,linebuf,BUFLEN);
	    if (i==BUFLEN && strcmp(linebuf,s)) j += (long)BUFLEN;
	    else break;
	}
	if (i==BUFLEN) {
	    lseek(namesf,j,0);
	    strcpy(d,&linebuf[MAXNAMELEN]);
	    return(1);
	}
    }
    return(0);
}

/* __________________________________________________________________
**|                                                                  |
**| patmat(s,p,insens) - simple regular expression pattern matcher   |
**|                                                                  |
**| Returns TRUE if string 's' fits regular expression pattern 'p'.  |
**| Regular expression pattern string 'p' can contain any of the     |
**| following:                                                       |
**|                                                                  |
**| ?		Matches any single character (exactly one char)      |
**| *		Matches any string of any length (0 or more chars)   |
**| \   	Match next character exactly                         |
**| [a-z]	Matches one char in the range "a-z"                  |
**| [~a-z]	Matches one char outside the range "a-z"             |
**| [^a-z]	Alternate form of [~a-z]                             |
**|                                                                  |
**| If 'insens' is specified TRUE, one-for-one character matches     |
**| are performed with case insensitivity (range matches are         |
**| always case sensitive).                                          |
**|__________________________________________________________________|
*/

#define ucase(x)   ((((x)>='a') && ((x)<='z')) ? (x-32) : (x))

patmat(s,p,insens)
char *s,*p;
int insens;
{
    int neg,matched;
    unsigned char last;
    while (1) {
	if (!*p) return(*s=='\0');
	switch (*p) {
	case '?': if (!*s) return(0); break;
	case '*': p++; return(*p ? stars(s,p,insens) : 1);
	case '[':
	    last=0xFF; matched=0; if (neg=((p[1]=='~')||(p[1]=='^'))) p++;
	    while (*++p && (*p != ']')) {
		if (*p=='-') { if (*s<=*++p && *s>=last) matched=1; }
		else if (*p==*s) matched=1;
		last = *p;
	    }
	    if (matched==neg) return(0); break;
	case '\\': p++; /* then drop into default case */
	default:
	    if (insens) { if (ucase(*s)!=ucase(*p)) return(0); }
	    else if (*s!=*p) return(0);
	}
	s++; p++;
    }
}

stars(s,p,insens)
char *s,*p;
int insens;
{
    do { if (patmat(s,p,insens)) return(1); } while (*++s);
    return(0);
}
