#ifndef lint
static char *RCSid = "$Header: assa.c,v 1.1 89/12/14 08:19:12 jay Exp $";
#endif /* not lint */

/*
 * assa.c
 *
 * hit the beaches!
 *
 * from PSL Empire, 1985
 *
 * assault <sector> <ship/tank/balloon#> <troops> [tanklist]
 * 
 * If tank numbers are optionally added, and the assaulting vessel
 * has landing capability, and the troop assault manages to capture
 * the beach, the specified tanks are also unloaded onto the beach.
 */

#include "misc.h"
#include "var.h"
#include "ship.h"
#include "sect.h"
#include "news.h"
#include "nat.h"
#include "file.h"
#include "xy.h"
#include "treaty.h"
#include "nsc.h"
#include "deity.h"

static	lostship();
char *STB();

assa()
{
	extern	char *argp[];
	extern	int btused;
	register int troops;
	struct	sctstr sect;
	struct	sctstr from;
	struct	shpstr ship,tank;
	struct	nstr_item nbst;
	coord	x, y;
	coord	shx, shy;
	int	shnu;
	natid	own;
	int	sdef;
	int	fdef;
	int	a_mil;
	int	s_mil;
	int	a_dead;
	int	d_dead;
	int	defenders;
	double	def;
	int	taken;
	double	odds;
	double	food;
	int	vec[I_MAX+1];
	int	pvec[I_MAX+1];
	int	svec[I_MAX+1];
	char	*p;
	float	food_per_trooper;

	if (!(p = getstarg(argp[1], "Sector :  ")) ||
	    !sarg_xy(p, &x, &y) || !getsect(x, y, &sect))
		return RET_SYN;

	shnu = onearg(argp[2],fmt("Assault sector %s from %s #:",
				  xyas(x,y,cnum),STB(0)));
	if (shnu < 0) {
	    pr(fmt("You may only assault from one %s!\n",
		   (strlen(STB(0))>4) ? "vehicle" : "ship"));
	    return RET_SYN;
	}
	if (!getship(shnu, &ship) || ship.shp_own != cnum) {
		pr("Not your vehicle.\n");
		return RET_FAIL;
	}
	shx = ship.shp_x;
	shy = ship.shp_y;
	if (mapdist(shx, shy, x, y) > 1) {
		pr("You'll have to get there first...\n");
		return RET_FAIL;
	}
	own = sect.sct_own;
	if (own == cnum) {
		pr("That's your own sector!\n");
		return RET_FAIL;
	}
	if (sect.sct_type <= SCT_WASTE) {
		pr("Don't bother...\n");
		return RET_FAIL;
	}
	getvec(VT_ITEM, vec, (char *)&ship, EF_SHIP);
	if (!(a_mil = vec[I_MILIT])) {
	    pr(fmt("No military on %s #%d",
		   mchr[ship.shp_type].m_name,shnu));
	    return RET_FAIL;
	}
	if (x != shx || y != shy) {
	    switch (vehicleclass(&ship)) {
	    case CLASS_SHIP:
		if ((!getsect(shx, shy, &from) ||
		     ((from.sct_type != SCT_WATER) &&
		      (from.sct_type != SCT_BSPAN)))) {
		    pr("Ship assaults can only be done while at sea!\n");
		    return RET_FAIL;
		}
		break;
	    case CLASS_TANK:
		if (!getsect(shx,shy,&from) ||
		    (from.sct_type == SCT_BSPAN)) {
		    pr(fmt("Your %s can't assault from a bridge span!\n",
			   mchr[ship.shp_type].m_name));
		    return RET_FAIL;
		}
		break;
	    case CLASS_BALLOON:
		pr(fmt("Your %s can only assault the sector it is over!\n",
		       mchr[ship.shp_type].m_name));
		return RET_FAIL;
	    case CLASS_TRAIN:
		pr(fmt("You can't assault from a %s!\n",
		       class[vehicleclass(&ship)].name));
		return RET_FAIL;
	    }
	}
	if (own && !trechk(cnum, own, LANATT))
		return RET_FAIL;
	getvec(VT_ITEM, svec, (char *)&sect, EF_SECTOR);
	s_mil = svec[I_MILIT];
	pr(fmt("Sector %s is a %d%% %s with approx. %d military.\n",
		xyas(x, y, cnum), round((int)sect.sct_effic, 10),
		dchr[sect.sct_type].d_name,
		round(s_mil, 10)));
	setbigmap(x, y, dchr[sect.sct_type].d_mnem);
	writebigmap();
	if (s_mil && !(mchr[ship.shp_type].m_flags & M_LAND))
		a_mil /= 10;
	troops = onearg(argp[3],fmt("Number of troops in assault? (max %d) ",
		a_mil));
	troops = min(troops, a_mil);
	if (troops <= 0)
		return RET_FAIL;
	/*
	 * NOTE: need getcom here
	 */
	food = vec[I_FOOD];
	taken = (int) (food * troops / (vec[I_MILIT] + vec[I_CIVIL]) + 0.5);
	vec[I_MILIT] -= troops;
	vec[I_FOOD] -= taken;
	putvec(VT_ITEM, vec, (char *)&ship, EF_SHIP);
	/* food_per_trooper = (float) food / troops; */
	food_per_trooper = (float) taken / troops;
	def = (dchr[sect.sct_type].d_dstr / 2.0 - 1.0) *
		(sect.sct_effic / 100.0) + 1.0;
	odds = (double) troops / (s_mil * def * 2 + troops);
	pr(fmt("Your success odds are %.1f%%\n", odds * 100.0));
	sdef = seadef(ship.shp_type);
	fdef = fortdef(&sect, sdef, shx, shy);
	if (fdef > 0) {
		shipdamage(&ship, fdef);
		putship(shnu, &ship);
		if (ship.shp_own == 0) {
			lostship(x, y, cnum, own, ship);
			return RET_OK;
		}
		troops = ((100 - fdef) * troops) / 100.0 + 0.5;
		odds = (double) troops / (s_mil * 2 + troops);
		pr(fmt("Your success odds are now %.1f%%\n", odds * 100.0));
	}
	fdef = defdef(own, sdef, shx, shy);
	if (fdef > 0) {
		wu(0, own, 
			fmt("Fortress(es) defending %s did %d%% damage to %s #%d",
			xyas(sect.sct_x, sect.sct_y, own), fdef,
			mchr[ship.shp_type].m_name, shnu));
		shipdamage(&ship, fdef);
		putship(shnu, &ship);
		if (ship.shp_own == 0) {
			lostship(x, y, cnum, own, ship);
			return RET_OK;
		}
		troops = ((100 - fdef) * troops) / 100.0 + 0.5;
		odds = (double) troops / (s_mil * 2 + troops);
		pr(fmt("Your success odds are now %.1f%%\n", odds * 100.0));
	}
	fdef = shipdef(own, sdef, shx, shy);
	if (fdef > 0) {
		wu(0, own, fmt("Return fire did %d%% damage to %s #%d",
			fdef, mchr[ship.shp_type].m_name, shnu));
		shipdamage(&ship, fdef);
		putship(shnu, &ship);
		if (ship.shp_own == 0) {
			lostship(x, y, cnum, own, ship);
			return RET_OK;
		}
		troops = ((100 - fdef) * troops) / 100.0 + 0.5;
		odds = (double) troops / (s_mil * 2 + troops);
		pr(fmt("Your success odds are now %.1f%%\n", odds * 100.0));
	}
	if (troops == 0) {
		lostship(x,y,cnum,own,ship);
		return RET_OK;
	}
	defenders = s_mil;
	a_dead = d_dead = 0;
	while (defenders > 0) {
		if (troops <= 0)
			break;
		if (chance(odds) == 0) {
			pr("@");
			a_dead++;
			troops--;
		} else {
			pr("!");
			d_dead++;
			defenders--;
		}
	}
	if (own != 0)
		wu(0, own,
	        fmt("%s (#%d) assaulted %s%s from %s #%d; yours %d, theirs %d\n",
		    cname(cnum), cnum, troops ? "(and took) " : " ",
		    xyas(x, y, own), mchr[ship.shp_type].m_name,
		    ship.shp_uid, d_dead, a_dead));
	if (troops > 0) {
		pr(fmt("\nYou have taken sector %s\n", xyas(x,y,cnum)));
		if (istank(&ship)) 
		    nreport(cnum, N_WON_SECT_TNK, own, 1);
		else if (isballoon(&ship))
		    nreport(cnum, N_WON_SECT_BAL, own, 1);
		else
		    nreport(cnum, N_WON_SECT, own, 1);
		troops = takeover(&sect, troops);
		sect.sct_mobil = 0;
		svec[I_MILIT] = troops;
		svec[I_FOOD] += roundavg(troops * food_per_trooper);
		if (sect.sct_type == SCT_CAPIT)
			caploss(&sect, own,
				"which happens to be %s's capital!\n");
	} else {
		pr("\nYou have been defeated!\n");
		if (istank(&ship))
		    nreport(cnum, N_SCT_LOSE_TNK, own, 1);
		else if (isballoon(&ship))
		    nreport(cnum, N_SCT_LOSE_BAL, own, 1);
		else
		    nreport(cnum, N_SCT_LOSE, own, 1);
		svec[I_MILIT] = defenders;
	}
	putvec(VT_ITEM, svec, (char *)&sect, EF_SECTOR);
	/* You've touched 'em -- give 'em plague if even you've lost! */
	getvec(VT_COND, vec, (char *)&ship, EF_SHIP);
	if (vec[C_PSTAGE] == PLG_INFECT) {
		getvec(VT_COND, pvec, (char *)&sect, EF_SECTOR);
		if (pvec[C_PSTAGE] == PLG_HEALTHY) {
			pvec[C_PSTAGE] = PLG_EXPOSED;
			putvec(VT_COND, pvec, (char *)&sect, EF_SECTOR);
		}
	}
	
	pr(fmt("Casualties :\nYours... %3d\nTheirs.. %3d\n", a_dead, d_dead));

	/*
        **  If we've secured a beach using a landing ship
        **  and now want to land tanks, do so now.
	**  Each tank we unload costs 1 btu.
	*/

	if ( troops>0 && isship(&ship) && ship.shp_ntank>0 &&
	    (mchr[ship.shp_type].m_flags & M_LAND) ) {
	    snxtitem(&nbst,EF_SHIP,"*");
	    pr("Landable vehicles:\n");
	    while (nxtitem(&nbst, (char *)&tank))
		if (istank(&tank) && 
		    tank.shp_own==cnum && tank.shp_aboard==shnu) {
		    pr(fmt("  %s #%d (%d%%, %d mob)\n",
			   mchr[tank.shp_type].m_name,tank.shp_uid,
			   tank.shp_effic,tank.shp_mobil));
		}
	    if (p=getstarg(argp[4],"Unload which tanks (* for all)? ")) {
		if (snxtitem(&nbst, EF_SHIP, p)) {
		    troops = 0;
		    while (nxtitem(&nbst, (char *)&tank)) {
			if (!(god || tank.shp_own==cnum)) continue;
			if (tank.shp_aboard != shnu) continue;
			if (!istank(&tank)) continue;
			if (ship.shp_ntank) ship.shp_ntank--;
			else ship.shp_ntank=0;
			tank.shp_aboard = -1;
			tank.shp_x = x;
			tank.shp_y = y;
			troops++;
			putship(tank.shp_uid,&tank);
			pr(fmt("Landing %s #%d\n",
			       mchr[tank.shp_type].m_name,tank.shp_uid));
		    }
		    pr(fmt("%d tank%s landed, %d remain%s aboard ship\n",
			   troops,(troops==1)?"":"s",
			   ship.shp_ntank,(ship.shp_ntank==1)?"s":""));
		    nreport(cnum, N_TNK_LAND, own, 1);
		    btused += troops;
		}
	    }
	}

	putsect(&sect);
	putship(shnu, &ship);
#ifdef ASSAULTBTUS
	btused += (int) ((d_dead + a_dead) * 0.15 + 0.5);
#else
	btused += (int) ((a_dead)*0.15 + 0.5);
#endif
	return RET_OK;
}

static
lostship(x, y, ctrynum, own, ship)
	coord	x;
	coord	y;
	natid	ctrynum;
	natid	own;
	struct	shpstr ship;
{
	char	*p;

	p = fmt("%s (#%d) assaulted %s from %s #%d", cname(ctrynum), ctrynum,
		xyas(x, y, own), mchr[ship.shp_type].m_name, ship.shp_uid);
	wu(0, own, p);
	nreport(ctrynum, N_ASSAULT_FAIL, own, 1);
}
