#ifndef lint
static char *RCSid = "$Header: boar.c,v 1.1 89/12/14 08:19:15 jay Exp $";
#endif /* not lint */

/*
 * boar.c
 *
 * board an enemy ship
 *
 * from PSL Empire, 1985
 */

#include "misc.h"
#include "var.h"
#include "nat.h"
#include "ship.h"
#include "news.h"
#include "file.h"
#include "treaty.h"
#include "loc.h"

char *STB();

boar()
{
	extern	char *argp[];
	int	btused;
	int	dshpno;		/* defending shipno */
	int	ashpno;		/* attacking shipno */
	int	amil;		/* # attacking mil */
	int	dmil;		/* # defending mil */
	int	smil;		/* # nonattacking mil */
	int	dmovable;	/* defender can move */
	int	milfromland;	/* mil boarding from land */
	natid	d_cnum;		/* defender cnum */
	coord	a_x;		/* attacker x */
	coord	a_y;		/* attacker y */
	coord	d_x;		/* defender x */
	coord	d_y;		/* defender y */
	int	acas;		/* attacker casualties */
	int	dcas;		/* defender casualties */
	int	dam;		/* damage from shipdef() */
	struct	shpstr aship;	/* attacking ship */
	struct	shpstr dship;	/* defending ship */
	double	aspeed;
	double	dspeed;
	char	*ptr;

	if (!(ptr = getstarg(argp[1],fmt("Victim %s #? ",STB(0)))))
	    return RET_SYN;

	if (milfromland = (int)index(ptr,',')) {
	    pr("Boarding from land not yet implemented\n");
	    return RET_FAIL;
	}

	if ((dshpno = atoi(ptr)) < 0 || !getship(dshpno, &dship)) {
		pr(fmt("No such %s\n",strlen(STB(0))>4 ? "vehicle" : "ship"));
		return RET_FAIL;
	}
	d_cnum = dship.shp_own;

	if (!trechk(cnum, d_cnum, SEAATT))
		return RET_FAIL;
	d_x = dship.shp_x;
	d_y = dship.shp_y;

	if (!(ptr=getstarg(argp[2],fmt("Boarding party from %s #? ",STB(0)))))
	    return RET_SYN;
	if ((ashpno = atoi(ptr)) < 0 || !getship(ashpno, &aship) ||
	    aship.shp_own != cnum)
		return RET_FAIL;
	a_x = aship.shp_x;
	a_y = aship.shp_y;
	if (a_x != d_x || a_y != d_y) {
		pr("Not in same sector!\n");
		return RET_FAIL;
	}

	/*
	**  Boarding vessel must contain some military.
	*/

	if ((amil = getvar(V_MILIT, (char *)&aship, EF_SHIP)) == 0) {
		pr(fmt("No military on %s\n", shipas(&aship,0)));
		return RET_FAIL;
	}

	/*
	**  Boarding vessel must have some mobility.
	*/

	if (aship.shp_mobil <= 0) {
		pr(fmt("%s has no mobility!\n", shipas(&aship,0)));
		return RET_FAIL;
	}

	/*
	**  Boarding vessel must be on the surface.
	*/

	if (aship.shp_depth > 0) {
	    pr(fmt("%s is %s!\n", shipas(&aship,0),
		   isballoon(&aship) ? "airborne" : "submerged"));
	    return RET_FAIL;
	}

	/*
	**  Vessel being boarded must be on the surface, too.
	*/

	if (dship.shp_depth > 0) {
	    pr(fmt("Can't %s %s.\n",
		   isballoon(&dship) ? "reach" : "see",
		   shipas(&dship,0)));
	    return RET_FAIL;
	}

	/*
	**  Tanks cannot be boarded if they are manned
	**  and fueled.
	*/

	if (istank(&dship)) {
	    if (getvar(V_MILIT, (char *)&dship, EF_SHIP) &&
		getvar(V_PETROL, (char *)&dship, EF_SHIP)) {
		pr(fmt("%s eludes your boarding attempt!\n",
		       shipas(&dship,0)));
		return RET_OK;
	    }
	}

	/*
	**  Ships in the neighborhood fire at the boarder.
	*/

	if ((dam = shipdef(d_cnum, seadef(aship.shp_type), a_x, a_y)) > 0) {
	    wu(0, d_cnum, fmt("Return fire did %d%% damage to %s #%d",
			      dam, mchr[aship.shp_type].m_name, ashpno));
	    shipdamage(&aship, dam);
	    putship(ashpno, &aship);
	    if (aship.shp_effic < 20) {
		wu(0, d_cnum,
		   fmt("Country #%d (%s) lost %s #%d trying to board %s",
		       cnum,cname(cnum), mchr[aship.shp_type].m_name, ashpno,
		       shipasc(&dship,LS,d_cnum)));
		nreport(cnum, N_SHP_LOSE, d_cnum, 1);
		return RET_OK;
	    }
	}

	/*
	**  Figure out of boarder can catch the victim.
	**  The victim ship can't evade without mobility, crew, and petrol.
	*/

	getship(ashpno, &aship);
	amil = getvar(V_MILIT, (char *)&aship, EF_SHIP);

	getship(dshpno, &dship);
	dmil = getvar(V_MILIT, (char *)&dship, EF_SHIP);

	dmovable = mchr[dship.shp_type].m_mpg;
	dmovable = (dmovable==0) ||
	    ((dmovable>0) && getvar(V_PETROL,(char *)&dship,EF_SHIP)>0);
	
	if (dmovable && dmil>0 && dship.shp_mobil>0) {
	    aspeed = mchr[aship.shp_type].m_speed *
		aship.shp_effic / 100.0;
	    dspeed = mchr[dship.shp_type].m_speed * 
		dship.shp_effic / 100.0;
	    aspeed = techfact(aship.shp_tech, 1.0) * aspeed;
	    dspeed = techfact(dship.shp_tech, 1.0) * dspeed;
	    aship.shp_mobil -= (int) (127.0 / aspeed);
	    dship.shp_mobil -= (int) (127.0 / aspeed);
	    if (aspeed < dspeed) {
		pr(fmt("Victim %s moves faster than you do!",
		       class[mchr[dship.shp_type].m_class].name));
		pr("  You can't catch him!\n");
		wu(0, dship.shp_own,
		   fmt("%s (#%d) %s #%d failed to catch %s #%d\n",
		       cname(aship.shp_own), aship.shp_own,
		       mchr[aship.shp_type].m_name, ashpno,
		       mchr[dship.shp_type].m_name, dshpno));
		putship(ashpno, &aship);
		putship(dshpno, &dship);
		return RET_OK;
	    }
	}

	/*
	**  Everything looks good, so do hand-to-hand combat!
	*/

	smil = amil;
	if (!(ptr=getstarg(argp[3],
			   fmt("Number of mil in boarding party (max %d)? ",
			       amil))))
	    return RET_SYN;
	if (*ptr = '.') acas = amil;
	else if ((acas = atoi(ptr)) <= 0) return RET_FAIL;
	if (acas > amil) acas = amil;
	amil = acas;
	smil -= amil;

	pr(fmt("%d attempting to board %s #%d\n", 
	       amil,mchr[dship.shp_type].m_name,dshpno));

	dcas = dmil;
	while (1) {
	    if (chance(0.33) || dmil <= 0) {
		pr("!");
		if (--dmil <= 0) break;
	    } else { 
		pr("@");
		if (--amil <= 0) break;
	    }
	}

	acas -= amil;
	dcas -= dmil;

	if (amil > 0) {
	    struct mchrstr *vbase;
	    pr("Boarding successful!\n");
	    /*
	    **  If anyone was aboard the victim ship, they manage
	    **  to get out a telegram ... Otherwise, the victor just
	    **  took an unmanned ship (and no one else will know).
	    */
	    if (dcas > 0) {
		wu(0, d_cnum, fmt("Country #%d (%s) boarded %s #%d\n",
				  cnum,cname(cnum),
				  mchr[dship.shp_type].m_name, dshpno));
		switch (vehicleclass(&dship)) {
		case CLASS_TANK:
		    nreport(cnum, N_BOARD_TANK, d_cnum, 1); break;
		case CLASS_BALLOON:
		    nreport(cnum, N_BOARD_BALLOON, d_cnum, 1); break;
		case CLASS_TRAIN:
		    nreport(cnum, N_BOARD_TRAIN, d_cnum, 1); break;
		default:
		    nreport(cnum, N_BOARD_SHIP, d_cnum, 1);
		}
	    }
	    takeover_ship(&dship, cnum);
	    dship.shp_fleet = ' ';
	    vbase = &mchr[dship.shp_type];
	    dmil = min(amil / 2, vl_find(V_MILIT, vbase->m_vtype,
					 vbase->m_vamt, (int) vbase->m_nv));
	    amil -= dmil;
	    pr(fmt("%d of your troops hold%s ship %d\n", dmil,
		   splur(dmil != 1), dshpno));
	    pr(fmt("%d return%s to ship %d\n", amil, splur(amil != 1),
		   ashpno));
	    if (getvar(V_PSTAGE, (char *)&dship, EF_SHIP) == PLG_INFECT) {
		if (getvar(V_PSTAGE, (char *)&aship, EF_SHIP) == 0)
		    putvar(V_PSTAGE, PLG_EXPOSED, (char *)&aship,
			   EF_SHIP);
	    }
	    else if (getvar(V_PSTAGE, (char *)&aship, EF_SHIP) ==
		     PLG_INFECT) {
		if (getvar(V_PSTAGE, (char *)&dship, EF_SHIP) == 0)
		    putvar(V_PSTAGE, PLG_EXPOSED, (char *)&dship,
			   EF_SHIP);
	    }
	} else {
	    pr("\nYou have been repelled...\n");
	    wu(0, d_cnum,
	       fmt("%s %s #%d tried to board %s #%d",
		   cname(cnum), mchr[aship.shp_type].m_name, ashpno,
		   mchr[dship.shp_type].m_name, dshpno));
	    switch (vehicleclass(&dship)) {
	    case CLASS_TRAIN: nreport(cnum, N_TRN_LOSE, d_cnum, 1); break;
	    case CLASS_TANK: /* not newsworth */ break;
	    case CLASS_BALLOON: nreport(cnum, N_BAL_LOSE, d_cnum, 1); break;
	    default: nreport(cnum, N_SHP_LOSE, d_cnum, 1); break;
	    }
	}
	putvar(V_MILIT, smil+amil, (char *)&aship, EF_SHIP);
	putship(ashpno, &aship);
	putvar(V_MILIT, dmil, (char *)&dship, EF_SHIP);
	putship(dshpno, &dship);
	pr(fmt("Casualties :\nYours ... %d\n", acas));
	pr(fmt("Theirs .. %d\n", dcas));
	btused = roundavg((dcas + acas) * 0.15 + 0.5);
	pr(fmt("Papershuffling ... %d B.T.U\n", btused));
	NAT_DELTA(nat_btu, cnum, -btused);
	return RET_OK;
}
