#ifndef lint
static char *RCSid = "$Header: laun.c,v 1.1 89/12/14 08:20:02 jay Exp $";
#endif /* not lint */

/*
 * laun.c
 *
 * Launch missiles from land or sea.
 *
 * Dave Pare, 1986
 */

#include "misc.h"
#include "xy.h"
#include "var.h"
#include "sect.h"
#include "item.h"
#include "plane.h"
#include "nuke.h"
#include "news.h"
#include "ship.h"
#include "nsc.h"
#include "nat.h"
#include "path.h"
#include "file.h"


/*
 * These are needed by the various launch routines below.
 */

extern  int etu_per_update;	/* etus/update */
extern  int orbit_rate;		/* basic orbit velocity, etus/sector */

static	struct	plnstr plane;	/* current missile we're lauinching */
static	struct	plchrstr *pcp;	/* characteristics of this missile */
static	int	shiplaunch;	/* launched from a ship? */
static	struct	sctstr sect;	/* sector from which we're launching */
static	struct	shpstr ship;	/* ship from which we're launching */

static	int launch_nuke();
static	int launch_sat();
static	int launch_conv();

/*
 * laun <PLANES>
 */
laun()
{
	extern	char *argp[];
	struct	nstr_item nstr;
	int n,ok;

	if (!snxtitem(&nstr, EF_PLANE, argp[1]))
		return RET_SYN;
	while (nxtitem(&nstr, (char *)&plane)) {
		if (plane.pln_own != cnum)
			continue;
		pcp = &plchr[plane.pln_type];
		if ((pcp->pl_flags & (P_M|P_O)) == 0) {
		    pr(fmt("%s #%d isn't a missile!\n", pcp->pl_name,
			   plane.pln_uid));
		    continue;
		}
 		if (pcp->pl_flags & P_F) {
		    pr(fmt("%s #%d is a surface-to-air missile!\n",
			   pcp->pl_name, plane.pln_uid));
		    continue;
 		}
		if (plane.pln_flags & PLN_BOXED) {
		    pr(fmt("%s #%d needs to be uncrated first!\n",
			   pcp->pl_name,plane.pln_uid));
		    continue;
		}
		if ((plane.pln_flags & PLN_LAUNCHED) && (pcp->pl_flags & P_O)) {
			pr(fmt("%s #%d already in orbit!\n",
				pcp->pl_name, plane.pln_uid));
			continue;
		}
		if (plane.pln_ship >= 0) {
			getship(plane.pln_ship, &ship);
			shiplaunch = 1;
			if (ship.shp_own != cnum){
				continue;
			}
			plane.pln_x = ship.shp_x;
			plane.pln_y = ship.shp_y;
		} else {
			shiplaunch = 0;
			getsect(plane.pln_x, plane.pln_y, &sect);
			if (sect.sct_own && sect.sct_own != cnum) {
			    ok = 0;
#ifdef TRADECENTERISAIRPORT
			    if (sect.sct_type==SCT_TRADE) ok = 1;
#endif			    
			    if (!ok) {
				pr(fmt("%s #%d: launch site %s not yours!\n",
				       pcp->pl_name, plane.pln_uid,
				       xyas(plane.pln_x, plane.pln_y, cnum)));
				continue;
			    }
			}
		}
		if (plane.pln_effic < 60) {
			pr(fmt("%s #%d is damaged (%d%%)\n",
				pcp->pl_name, plane.pln_uid, plane.pln_effic));
			continue;
		}
		if (plane.pln_nukeamt == 0 &&
		    (pcp->pl_flags & (P_M|P_O)) == (P_M|P_O)) {
			pr(fmt("%s #%d isn't armed with a nuclear device!\n",
				pcp->pl_name, plane.pln_uid));
			continue;
		}
		pr(fmt("%s #%d at %s; range %d, eff %d%%", pcp->pl_name,
			plane.pln_uid,
			xyas(plane.pln_x, plane.pln_y, cnum),
			plane.pln_range,
			plane.pln_effic));
		n = plane.pln_nukeamt;
		if (!(pcp->pl_flags & P_O) && (n == 0)) { /* conventional */
			if (launch_conv() < 0)
				continue;
		} else if (n == 0) {			/* satellites */
			if (launch_sat() < 0)
				continue;
		} else {				/* nukes */
			if (launch_nuke(n) < 0)
				continue;
		}
		putplane(plane.pln_uid, &plane);
		if (shiplaunch) {
			ship.shp_nplane -= 1;
			if (ship.shp_nplane < 0)	/* shouldn't happen */
				ship.shp_nplane = 0;
			putship(plane.pln_ship, &ship);
		}
	}
	return RET_OK;
}

/*
**  Launch a conventional warhead missile.
**  Return -1 on failure, 0 on success (even if missile explodes).
*/

static int
launch_conv()
{
    coord sx, sy;
    float damage;
    int	dam;
    int	shell;
    int	bombs;
    int	needed;
    int	i;
    int targetid;
    struct shpstr targetship;
    char *cp,*s;
    extern char *argp[];
    pr(", carrying conventional warhead\n");
    if (shiplaunch) {
	shell = getvar(V_SHELL, (char *)&ship, EF_SHIP);
    } else {
	shell = getvar(V_SHELL, (char *)&sect, EF_SECTOR);
    }
    needed = pcp->pl_load - 1;
    if (shell < needed) {
	pr(fmt("(#%d) %s not enough shells!\n",
	       plane.pln_uid, pcp->pl_name));
	return -1;
    }
    bombs = pcp->pl_load;
    if (shiplaunch) {
	putvar(V_SHELL, shell - needed, (char *)&ship, EF_SHIP);
    } else {
	putvar(V_SHELL, shell - needed, (char *)&sect, EF_SECTOR);
    }
    cp = argp[2];
    while (1) {
	cp = getstarg(cp, "Target sector? ");
	if (!cp || !*cp) return (-1);
	if (!sarg_xy(cp, &sx, &sy)) {
	    pr("Bad sector designation; try again!\n");
	    cp = 0;
	    continue;
	}
	if (mapdist(plane.pln_x,plane.pln_y,sx,sy) > plane.pln_range) {
	    pr("Range too great; try again!\n");
	    cp = 0;
	    continue;
	}
	targetid = -1;
	if (pcp->pl_flags & P_T) {
	    cp = argp[3];
	    if (cp=getstarg(cp, "Tactical target id? ")) {
		if (isdigit(*cp) && !index(cp,',')) {
		    targetid = atoi(cp);
		    if (getship(targetid,&targetship))
			if (targetship.shp_own) break;
		    pr("Bad ship id; try again!\n");
		    cp = 0;
		    continue;
		}
	    }
	}
	break;
    }
    pr("Launching!\n");
    if (pcp->pl_flags & P_T)
	pr("Ffwwwoooossssshhhhhh!\n"); /* realistic cruise missile noise */
    else if (chance(0.02 + (100 - plane.pln_effic)/100.0)) {
	pr("KABOOOOM!  Missile explodes on launch pad!\n");
	return 0;
    }

    if (pcp->pl_flags & P_T)
	if (!flycruisemissile(&plane,sx,sy)) return 0;

    if (targetid >= 0) {
	if (chance(pcp->pl_acc*pcp->pl_tech/plane.pln_tech/100.0) || 
	    targetship.shp_x!=sx || targetship.shp_y!=sy ||
	    (issub(&targetship) && targetship.shp_depth>0)) {
	    pr(fmt("Missile failed to lock onto tactical target #%d\n",
		   targetid));
	    targetid = -1;
	} else {
	    pr("Target lock!\n");
	}
    }
    dam = 100;
    i = roll(100);
    if (i <= 10)                  { /* 8% */ damage = 0.92; pr("BLAM"); } 
    else if (i < 100-pcp->pl_acc) { /* 5% */ damage = 0.95; pr("Blam"); }
    else                          { /* 2% */ damage = 0.98; pr("blam"); }
    pr("\n");
    for (i=0; i<bombs; i++) dam *= damage;
    dam = 100-dam;
    plane.pln_own = 0;
    if (targetid >= 0) {
	if (isship(&targetship)) 
	    nreport(cnum, N_MSL_HIT_SHIP, targetship.shp_own, 1);
	else if (istank(&targetship)) dam = min(dam*4,100);
	else if (isballoon(&targetship)) dam = 100;
	pr(fmt("did %d%% damage to %s %s #%d in %s\n",
	       dam,cname(targetship.shp_own),
	       mchr[targetship.shp_type].m_name,targetid,
	       xyas(sx,sy,cnum)));
	wu(0, targetship.shp_own,
	   fmt("%s missile attack did %d%% damage to %s #%d in %s\n",
	       cname(cnum),dam,
	       mchr[targetship.shp_type].m_name,targetid,
	       xyas(sx,sy,targetship.shp_own)));
	shipdamage(&targetship, dam);
	putship(targetid,&targetship);
    } else {
	nreport(cnum, N_SCT_BOMB, sect.sct_own, 1);
	pr(fmt("did %d%% damage in %s\n", dam, xyas(sx, sy, cnum)));
	wu(0, sect.sct_own,
	   fmt("%s missile attack did %d%% damage in %s",
	       cname(cnum), dam, xyas(sx, sy, sect.sct_own)));
	getsect(sx, sy, &sect);
	sectdamage(&sect, dam);
	putsect(&sect);
    }
    return 0;
}

/*
**  Launch a satellite.
**  Return -1 on error, 0 on success (even if the satellite fails).
*/

static int launch_sat()
{
    coord sx, sy;
    int	i,j,dist,dir,orbitvel,orbitdir;
    char *cp,*s;
    extern char *argp[];
 
    pr("\n");
    cp = argp[2];
    while (1) {
	cp = getstarg(cp,"Target sector? ");
	if (cp == 0) return -1;
	if (!sarg_xy(cp, &sx, &sy)) {
	    pr("Bad sector designation; try again!\n");
	    cp = 0;
	    continue;
	}
	if ((dist = mapdist(plane.pln_x, plane.pln_y, sx, sy)) >
	    plane.pln_range) {
	    pr("Range too great; try again!\n");
	    cp = 0;
	    continue;
	}
	break;
    }
    for (orbitdir = -1; orbitdir < 0; ) {
	cp = getstarg(argp[3],"Direction of orbit (CR=none)? ");
	if (!cp || !*cp) cp = "h";
	orbitdir = chkdir(*cp,0,6);
	if (orbitdir<0 || orbitdir>6) {
	    pr(fmt("`%c' not a valid orbit direction\n",*cp)); 
	    direrr("`%c' for stationary orbit",0,0);
	}
    }
    if (orbitdir) {
	for (orbitvel = -1; orbitvel < 0; ) {
	    cp = getstarg(argp[4],"Orbital velocity (1-4; 0 for help)? ");
	    if (!cp || !isdigit(*cp)) cp = "2";
	    i = atoi(cp);
	    if (i>0 && i<5) orbitvel = i-1;
	    else {
		pr("Must be a value from 1 to 4, indicating:\n");
		for (i=1; i<5; i++) {
		    j = (i * etu_per_update) / orbit_rate;
		    pr(fmt("  %d: movement of %d sectors/update\n",i,j));
		}
		pr("try again!\n");
	    }
	}
    }
    pr("3... 2... 1... Blastoff!!!\n");
    if (chance(0.07 + (100 - plane.pln_effic)/100.0)) {
	pr(fmt("KABOOOOM!  Range safety officer detonates booster!\n"));
	plane.pln_own = 0;
	return 0;
    }
    i = plane.pln_tech + plane.pln_effic;
    if (chance(1.0 - (i/(i+50.0)))) {
	dir = (random() % 6) + 1;
	sx += diroff[dir][0];
	sy += diroff[dir][1];
	pr("Your trajectory was a little off.\n");
    }
    nreport(cnum, N_LAUNCH, 0, 1);
    pr(fmt("%s #%d positioned over %s, ",
	   pcp->pl_name, plane.pln_uid, xyas(sx, sy, cnum)));
    plane.pln_x = xnorm(sx);
    plane.pln_y = ynorm(sy);
    plane.pln_flags |= PLN_LAUNCHED;
    plane.pln_flags |= (orbitdir & PLN_ORBITDIR);
    if (plane.pln_flags & PLN_ORBITDIR) {
	plane.pln_flags |= ((orbitvel << 3) & PLN_ORBITVEL);
	dpr(fmt("launch_sat: orbitvel=%d",orbitvel));
	dpr(fmt("launch_sat: plane.pln_flags=0x%x",plane.pln_flags));
	i = 1 + (plane.pln_flags & PLN_ORBITVEL)>>3;
	dpr(fmt("launch_sat: extracting: i=%d",i));
	i = (i * etu_per_update) / orbit_rate;
	pr(fmt("orbit vector %d%c,\n",i,
	       dirch[plane.pln_flags & PLN_ORBITDIR]));
    }
    plane.pln_mobil = (plane.pln_mobil > dist) ? 
	(plane.pln_mobil - dist) : 0;
    pr(fmt("will be ready for use in %d time units\n",
	   127 - plane.pln_mobil));
    return 0;
}

#define	MAXRV	30

struct target {
    int x, y;
    int airburst;
};

/*
**  Launch a nuke.
**  Return -1 on error, 0 on success (even if nuke fails).
*/

static int launch_nuke(n)
int n;
{
    coord sx, sy;
    struct target xy[MAXRV];
    char *p;
    int	i;
    int	dir;
    extern char *argp[];

    pr(fmt(", payload %dx%s\n", plane.pln_nukeamt,
	   nchr[plane.pln_nuketype].n_name));
    if (n > MAXRV) n = MAXRV;

    if (pcp->pl_flags & P_T) {
	p = argp[2];
	while (1) {
	    p = getstarg(p, "Target sector? ");
	    if (p == 0) return (-1);
	    if (!sarg_xy(p, &sx, &sy)) {
		pr("Bad sector designation; try again!\n");
		p = 0;
		continue;
	    }
	    if (mapdist(plane.pln_x,plane.pln_y,sx,sy) > plane.pln_range) {
		pr("Range too great; try again!\n");
		p = 0;
		continue;
	    }
	    break;
	}
	for (i=0; i<n; i++) {
	    xy[i].x = sx;
	    xy[i].y = sy;
	    xy[i].airburst = (pcp->pl_flags & P_O) ? 2 : 0;
	}
    }
    else for (i = 0; i < n;) {
	p = getstring(fmt("RV #%d, target sector? ", i+1));
	if (p == 0) return (-1);
	if (!sarg_xy(p, &sx, &sy)) {
	    pr("Bad sector designation; try again!\n");
	    continue;
	}
	if (mapdist(plane.pln_x,plane.pln_y,sx,sy) > plane.pln_range) {
	    pr("Range too great; try again!\n");
	    continue;
	}
	if (pcp->pl_flags & P_O) {
	    dir = plane.pln_tech + plane.pln_effic;
	    if (chance(1.0 - dir/(dir+50.0)))
		dir = (random() % 6) + 1;
	    else
		dir = 0;
	    xy[i].x = sx + diroff[dir][0];
	    xy[i].y = sy + diroff[dir][1];
	    xy[i].airburst = 2;
	} else {
	    p = getstring("Air-burst? ");
	    if (p == 0)
		return -1;
	    xy[i].x = sx;
	    xy[i].y = sy;
	    xy[i].airburst = 1;
	    if (*p == 0 || *p == 'n')
		xy[i].airburst = 0;
	}
	i++;
    }

    plane.pln_own = 0;
    plane.pln_nukeamt = 0;
    pr("Launching!\n");
    
    /*
    **  There is an [inefficiency+2] percent chance that the
    **  missile will explode.  There is a 25% chance that, if
    **  it explodes, it will explode in the air.  
    **  Also, if it explodes, there is a small percent chance,
    **  based on the tech level of the warheads, that the 
    **  warheads will detonate.
    */
    
    if (pcp->pl_flags & P_T)
	pr("Ffwwwoooossssshhhhhh!\n"); /* realistic cruise missile noise */
    else if (chance(0.02 + (100-plane.pln_effic)/100.0)) {
	float oops,m1,m2;
	int swore,intheair;
	if (intheair = chance(0.25))
	    pr(fmt("KABOOOOM!  Missile explodes seconds after launch!\n"));
	else
	    pr(fmt("KABOOOOM!  Missile explodes on launch pad!\n"));
	m1 = ((T250odds*350.0)-(T350odds*250.0))/(T250odds-T350odds);
	m2 = (m1-250.0)/T250odds;
	oops = (m1-(float)nchr[plane.pln_nuketype].n_tech)/m2;
	swore = 0;
	if (oops > 0.0) for (i=0; i<n; i++)
	    if (chance(oops)) {
		if (!swore) {
		    pr("Oh, sh*t!...\n");
		    swore = 1;
		}
		detonate(plane.pln_nuketype, xy[i].x, xy[i].y,
			 intheair,cnum);
	    }
	return 0;
    }
    
    pr("Arming warheads...\n");
    if (!(pcp->pl_flags & P_T))
	pr("Releasing RV's...\n");
    for (i = 0; i < n; i++) {
	detonate(plane.pln_nuketype, xy[i].x, xy[i].y,
		 xy[i].airburst, cnum);
    }
    return 0;
}

flycruisemissile(plane,sx,sy)
struct plnstr *plane;
coord sx,sy;
{
    /*
    **  Generate a routing path for cruise missiles.
    **  Countries along the way may may try to shoot it down.
    */

    char *mpath;
    struct qelem list,null;
    struct plchrstr *pcp;
    struct plist *plp;
    extern char *argp[];
    pcp = &plchr[plane->pln_type];
    plp = (struct plist *) malloc(sizeof(struct plist));
    plp->state    = P_OK;
    plp->misc     = 0;
    plp->bombs    = 0;
    plp->kamikaze = -1;
    plp->pcp      = pcp;

    bcopy((char *)plane, (char *)&plp->plane, sizeof(struct plnstr));
    initque(&list);
    initque(&null);
    insque(&plp->queue, &list);

    mpath = (char *)missilepath(plane->pln_x,plane->pln_y,sx,sy);

    pr(fmt("(#%d) %s launched\n", plane->pln_uid, pcp->pl_name));
    pr(fmt("Flightpath: %s\n",mpath));

    fly_flightpath(&list,&null,plane->pln_x,plane->pln_y,
		   mpath, (plane->pln_tech<<16)+P_X, &sx, &sy, 
		   plane->pln_effic);

    pln_put(&list);
    getplane(plane->pln_uid,plane);
    return ((plane->pln_own==cnum) && (plane->pln_effic >= 10));
}
