#ifndef lint
static char *RCSid = "$Header: paths.c,v 1.1 89/12/14 13:28:28 jay Exp $";
#endif

/*
 * path.c
 *
 * routines associated with paths, directions, etc.
 *
 * from PSL Empire, 1985
 */

#include <stdio.h>
#include "gamesdef.h"
#include "misc.h"
#include "xy.h"
#include "path.h"
#include "nat.h"
#include "sect.h"
#include "file.h"

#define MAXLEV 6
#define MAXBUF 256
#define MAXLEG 100

char pathnamefil[] = EMPPATH(data/paths);

int
getdir(prompt, stop_msg, view_msg, bomb_msg)
	char   *prompt;
	char   *stop_msg;
	char   *view_msg;
	char   *bomb_msg;
{
	register int max_dir;
	register int min_dir;
	register int dir_num;
	register char *buf;

	if (stop_msg != 0)
		min_dir = DIR_STOP;
	else
		min_dir = DIR_FIRST;
	if (view_msg == 0)
		max_dir = DIR_LAST;
	else
		max_dir = DIR_VIEW;
	while (1) {
		if ((buf = getstring(prompt)) == 0 || *buf == 0)
			return -1;
		dir_num = chkdir(buf[0], min_dir, max_dir);
		if (dir_num >= min_dir)
			break;
		direrr(stop_msg, view_msg, bomb_msg);
	}
	return dir_num;
}

int
chkdir(dir_char, min_dir, max_dir)
	char    dir_char;
	int     min_dir;
	int     max_dir;
{
	register int i;

	for (i = min_dir; i <= max_dir; i++)
		if (dir_char == dirch[i])
			return i;
	return -1;
}

direrr(stop_msg, view_msg, map_msg)
	char   *stop_msg;
	char   *view_msg;
	char   *map_msg;
{
	pr("Legal directions are:\n");
	pr(fmt(" %c %c\n", dirch[DIR_UL], dirch[DIR_UR]));
	pr(fmt("%c   %c\n", dirch[DIR_L], dirch[DIR_R]));
	pr(fmt(" %c %c\n", dirch[DIR_DL], dirch[DIR_DR]));
	if (stop_msg != 0)
		pr(fmt(stop_msg, dirch[DIR_STOP]));
	if (view_msg != 0)
		pr(fmt(view_msg, dirch[DIR_VIEW]));
	if (map_msg != 0)
		pr(fmt(map_msg, dirch[DIR_MAP]));
}

/*
 * return pointer to path; prompt user until a stop char
 * or a bomb char have been entered.  A "bomb" char in
 * this context is actually "execute" for the partial
 * move commands, and isn't valid for those commands
 * which do not accept partial moves.
 */
char *getpath(arg, x, y, onlyown, showdes, showxy)
int onlyown,showdes,showxy;
char *arg;
coord x,y;
{
    char	*pathbuf[MAXLEV]; /* input buffers (malloc'ed) */
    char	*pathptr[MAXLEV]; /* input buffer read pointers */
    char	alloced[MAXLEV];  /* allocation flags */
    static	char buf[MAXBUF]; /* output buffer (static) */

    char	*inp;		  /* current input pointer */
    char 	*outp;		  /* current output pointer */
    int		len;		  /* output buffer length */
    
    char   	terrain[32];	       /* terrain restriction string */
    char	tbuf[MAXBUF+MAXLEG+1];
    char	*s,*t,*u,*o=buf;
    char	*prompt;
    char	*execute;
    char	*p1;
    coord	nx, ny;
    int		dir,i,j,pathlevel;
    struct	sctstr sect;

    if (showxy) execute = " & '%c' to execute\n";
    else execute = "\n";
    
    inp = pathbuf[0] = arg ? arg : "";
    *buf = *terrain = '\0';
    len = pathlevel = 0;
    outp = buf;
    goto more;
    
 toodeep:
    
    pr("Path buffers nested too deep.\n");
    
 reset:
    
    while (pathlevel) {
	if (alloced[pathlevel]) free(pathbuf[pathlevel]);
	pathlevel--;
    }

    *inp = '\0';
    
 more:	
    while (*inp) {
	
	switch (*inp) {

	case '@':  /* GET NAMED PATH */

	    t = inp+1;
	    for (s=t; *t && *t!=';' && *t!='[' && *t!='?' && *t!='@'; t++);
	    s = (t==s) ? (char *) 0 : (char *) getnamedpath(s,t-s,cnum);
	    if (!s) {
		*t = '\0';
		pr("Invalid pathname \"%s\".\n",s);
		goto reset;
	    } 
	    if (pathlevel >= MAXLEV-1) goto toodeep;
	    pathptr[pathlevel++] = t; /* save this pathbuf's ptr */
	    pathptr[pathlevel] = s;   /* init new pathbuf's ptr */
	    alloced[pathlevel] = 1;   /* note that we malloc'ed it */
	    inp = s;
	    break;
	    
	case '[':  /* ROUTE RESTRICTION STRING */

	    t = inp+1;
	    for (s=terrain; *t && *t!=']' && (s-terrain)<31; *s++ = *t++);
	    *s = '\0';
	    if (*t && *t!=']') {
		pr("Autoroute terrain restrictions truncated.\n");
		while (*t && *t!=']') t++;
	    }
	    if (*t) t++;		
	    if (*terrain) 
		pr(fmt("Autorouting restricted to [%s] terrain.\n",
		       terrain));
	    else
		pr("Removing autoroute terrain restrictions.\n");
	    inp = t;
	    break;
	
	case '?':  /* SHOW PATH SO FAR */

	    pr(fmt("Path so far: %s\n",buf));
	    pr(fmt("Autoroute terrain: [%s]\n",terrain));
	    inp++;
	    break;
	
	case ';':  /* NOP (CONTINUE) */

	    inp++;
	    break;
	
	case 'd':  /* ERASE (BACKTRACK) */

	    if (!isdigit(*++inp)) i = 1;
	    else {
		i = atoi(inp);
		while (isdigit(*inp)) inp++;
	    }
	    if (i > len) i = len;
	    len -= i;
	    while (i--) {
		dir = chkdir(*--outp, DIR_STOP, DIR_LAST);
		if (dir >= 0) {
		    x -= diroff[dir][0];
		    y -= diroff[dir][1];
		} 
	    }
	    *outp = '\0';
	    break;
	
	default:

	    if (sarg_xy(inp,&nx,&ny)) { /* AUTOROUTE COORDINATES */
		int legaldest();
		int legalpath();
		legalroute(terrain,"",nx,ny);
		s = (char *)shortestpath(x,y,legaldest,legalpath,MAXLEG);
		if (!*s || *s=='?') {
		    pr(fmt("Can't autoroute from %s to %s.\n",
			   xyas(x,y,cnum),xyas(nx,ny,cnum)));
		    goto reset;
		}
		/* get rid of terminating 'h' */
		for (t=s; *t && *t!='h'; t++);
		if (*t=='h') *t='\0';
		if (strlen(s) > MAXLEG) {
		    pr(fmt("Path from %s to %s exceeds %d sectors.\n",
			   xyas(x,y,cnum),xyas(nx,ny,cnum),MAXLEG));
		    goto reset;
		}
		while (*inp && (isdigit(*inp)||*inp==','||*inp=='-')) inp++;
		if (pathlevel >= MAXLEV-1) goto toodeep;
		pathptr[pathlevel++] = inp; /* save this pathbuf's ptr */
		pathptr[pathlevel] = s;     /* init new pathbuf's ptr */
		alloced[pathlevel] = 0;     /* note it's not malloc'ed */
		if (*s) pr(fmt("%s to %s: %s\n",
			       xyas(x,y,cnum),xyas(nx,ny,cnum),s));
		inp = s;
	    }
	    
	    if (isdigit(*inp)) {
		pr("Invalid coordinate.\n");
		goto reset;
	    }
	    
	    dir = chkdir(*inp, DIR_STOP, DIR_LAST);
	    if (dir < 0) {
		pr(fmt("\"%c\" is not legal...", *inp));
		direrr("'%c' to stop", (char *)0, execute);
		goto reset;
	    }
	    
	    nx = x + diroff[dir][0];
	    ny = y + diroff[dir][1];
	    getsect(nx, ny, &sect);
	    
	    if (onlyown && sect.sct_own != cnum) {
		pr(fmt("You don't own %s; you can't go there!\n",
		       xyas(nx, ny, cnum)));
		goto reset;
	    }
	    
	    if (dir == DIR_STOP || dir == DIR_MAP) {
		dpr(fmt("stopping, returning buf = \"%s\"\n",buf));
		while (pathlevel) {
		    if (alloced[pathlevel]) free(pathbuf[pathlevel]);
		    pathlevel--;
		}
		return buf;
	    }
	     
	    if (len < MAXBUF-2) {
		*o++ = *inp++;
		*o = '\0';
		len++;
	    } else {
		pr("Path would exceed length limit.\n");
		goto reset;
	    }
	    
	    x = nx;
	    y = ny;
	    
	}
    }
    
    /*
    **  We get here ONLY if we've exhausted the current input level.
    */
    
    if (pathlevel) {
	if (alloced[pathlevel]) 
	    free(pathbuf[pathlevel]);
	pathlevel--;
	inp = pathptr[pathlevel];
	goto more;
    }
    
    /* 
    **  We get here ONLY if we've exhausted input from stdin 
    **  (and need more).
    */
    
    if (showdes) {
	getsect(x, y, &sect);
	prompt = fmt("<%c: %s> ",dchr[sect.sct_type].d_mnem,xyas(x, y, cnum));
    } else
	prompt = fmt("<%d: %s> ", len, xyas(x, y, cnum));
    
    if ((pathbuf[0] = getstring(prompt)) == 0 || *pathbuf[0] == 0) {
	dpr(fmt("null response, returning buf = \"%s\"\n",buf));
	return buf;
    }

    inp = pathbuf[0];
    goto more;
}



/* ARGSUSED */
double mcost(sp, own)
struct	sctstr *sp;
int	own;
{
	int	m_cost;

	if ((m_cost = dchr[sp->sct_type].d_mcst) == 0)
		return -1.0;
	return (m_cost * 100.0 - (double)sp->sct_effic) / 500.0;
}

/*
 * fly move cost
 */
/* ARGSUSED */
double
fcost(sp, own)
	struct	sctstr *sp;
	natid	own;
{
	return 1.0;
}

/*
 * nav move cost
 */
/* ARGSUSED */
double
ncost(sp, own)
	struct	sctstr *sp;
	natid	own;
{
	return 1.0;
}

/*
 * translate a 32 bit mask into a 10 element path
 */
char *
masktopath(mask)
	long	mask;
{
	static	char pathstr[32];
	int	i;
	int	val;
	char	*pp;

	pp = pathstr;
	for (i=0; i < sizeof(mask)*8; i += 3) {
		val = (mask & 0x7);
		if (val == 0)
			break;
		*pp++ = dirch[val];
		mask >>= 3;
	}
	*pp = '\0';
	return pathstr;
}

/*
 * translate a path into a bit mask...for now no longer
 * than 32 bits.
 */
long
pathtomask(path)
	char	*path;
{
	register char *pp;
	long	mask;
	int	i;
	int	val;

	mask = 0;
	for (i=0, pp=path; i < sizeof(mask)*8 && *pp; pp++, i += 3) {
		if ((val = chkdir(*pp, DIR_STOP, DIR_LAST)) == 0)
			break;
		mask |= val << i;
	}
	return mask;
}

/*
 * return end x,y of path, and the base
 * movement cost it takes to get there.
 */
double
pathtoxy(path, xp, yp, cost)
	char	*path;
	coord	*xp;
	coord	*yp;
	double	(*cost)();
{
	struct	sctstr s;
	char	*pp;
	coord	x;
	coord	y;
	int	val;
	double	m;
	int	c;

	x = *xp;
	y = *yp;
	m = 0.0;
	for (pp = path; *pp; pp++) {
		if ((val = chkdir(*pp, DIR_STOP, DIR_LAST)) == 0)
			break;
		x += diroff[val][0];
		y += diroff[val][1];
		c = dirch[val];
		if (c == DIR_STOP)
			break;
		if (!getsect(x, y, &s))
			return -1.0;
		m += cost(&s, s.sct_own);
	}
	*xp = x;
	*yp = y;
	return m;
}

/*
 * return true if "who" owns the path starting at x,y
 */
int
chkpath(who, path, x, y)
	natid	who;
	char	*path;
	coord	x;
	coord	y;
{
	char	*pp;
	int	val;
	struct	sctstr sect;

	for (pp = path; *pp; pp++) {
		if ((val = chkdir(*pp, DIR_STOP, DIR_LAST)) == 0)
			break;
		x += diroff[val][0];
		y += diroff[val][1];
		if (!getsect(x, y, &sect) || sect.sct_own != who)
			return 0;
	}
	return 1;
}

pathrange(cx, cy, pp, border, range)
	register coord	cx, cy;
	register char	*pp;
	int		border;
	struct	range	*range;
{
	int	dir;

	range->lx = cx;
	range->hx = cx;
	range->ly = cy;
	range->hy = cy;
	range->width = 0;
	range->height = 0;
	while ((dir = chkdir(*pp, DIR_FIRST, DIR_LAST)) >= 0) {
		pp++;
		cx += diroff[dir][0];
		cy += diroff[dir][1];
		if (cx < range->lx)
			range->lx = cx;
		if (cx > range->hx)
			range->hx = cx;
		if (cy < range->ly)
			range->ly = cy;
		if (cy > range->hy)
			range->hy = cy;
	}
	range->lx = xnorm(range->lx - border*2);
	range->ly = ynorm(range->ly - border);
	range->hx = xnorm(range->hx + border*2 + 1);
	range->hy = ynorm(range->hy + border + 1);
	xysize_range(range);
}

char *missilepath(x,y,tx,ty)
int x,y,tx,ty;
{
    int dxR,dxL,dyD,dyU,i,dir;
    static char pathbuf[200];

    tx = xnorm(tx);
    ty = ynorm(ty);
    x = xnorm(x);
    y = ynorm(y);

    dpr(fmt("missilepath: %d,%d -> %d,%d",x,y,tx,ty));

    for (i=0,dir=1; dir>0 && i<199; i++) {

	if (tx > x) {
	    dxR = tx-x;		/* distance going right */
	    dxL = x+WORLD_X-tx;	/* distance going left */
	} else if (tx < x) {
	    dxR = tx+WORLD_X-x;	/* distance going right */
	    dxL = x-tx;		/* distance going left */
	} else {
	    dxR = dxL = 0;
	}

	if (ty > y) {
	    dyD = ty-y;		/* distance going down */
	    dyU = y+WORLD_Y-ty;	/* distance going up */
	} else if (ty < y) {
	    dyD = ty+WORLD_Y-y;	/* distance going down */
	    dyU = y-ty;		/* distance going up */
	} else {
	    dyD = dyU = 0;	/* distance up/down = 0 */
	}

	if (dxR < dxL) {	
	    if (dyD < dyU)      dir=3;		/* go right & down */
	    else if (dyU < dyD) dir=1;		/* go right & up */
	    else                dir=2;		/* go straight right */
	} else if (dxL < dxR) { 
	    if (dyD < dyU)      dir=4;		/* go left & down */
	    else if (dyU < dyD) dir=6;		/* go left & up */
	    else                dir=5;		/* go straight left */
	} else if (dyD < dyU)   dir=roll(2)+2;	/* go somehow down */
	else if (dyU < dyD)     dir=roll(2)*5-4;/* go somehow up */
	else dir=0;
	
	pathbuf[i] = dirch[dir];

	dpr(fmt("  %d.  [dxR:dxL],[dyD:dyU] = [%d:%d],[%d:%d], dir=%d (%c)",
		i,dxR,dxL,dyD,dyU,dir,pathbuf[i]));

	x = xnorm(x+diroff[dir][0]);
	y = ynorm(y+diroff[dir][1]);

    }

    if (dir>0)
	*pathbuf='\0';
    else {
	if (pathbuf[i] != dirch[0])
	    pathbuf[i++] = dirch[0];
	pathbuf[i] = '\0';
    }
    return pathbuf;
}

/* ________________________________________________________________
**|                                                                |
**| Named Path Routines                                            |
**|________________________________________________________________|
*/

delnamedpath(name,cn)
char *name;
natid cn;
{
    int i;
    FILE *fout,*fin;
    char fil[128];
    char buf[1024];
    sprintf(fil,"%s%d",pathnamefil,cn);
    if (!(fin=fopen(fil,"r")))
	return RET_OK;
    strcat(fil,"#");
    if (!(fout=fopen(fil,"w"))) {
	logerror(fmt("delnamedpath(%s,%d): file creation error\n",name,cn));
	fclose(fin);
	return RET_FAIL;
    }
    i = strlen(name);
    while (fgets(buf,1024,fin))
	if (strncmp(buf,name,i) || buf[i]!=' ') 
	    fprintf(fout,buf);
    fclose(fout);
    fclose(fin);
    strcpy(buf,fil);
    buf[strlen(buf)-1] = '\0';
    link(fil,buf);
    unlink(fil);
    return RET_OK;
}

setnamedpath(name,path,cn)
char *name,*path;
natid cn;
{
    FILE *fout;
    char *s,*e;
    char buf[1024];
    if (s=(char *)getnamedpath(name,strlen(name),cn)) {
	free(s);
	pr(fmt("\"%s\" must be deleted before it can be redefined\n",name));
	return RET_FAIL;
    }
    for (s=name; *s && *s!=';' && *s!='[' && *s!='?' && *s!='@'; s++);
    if (*s) {
	pr(fmt("\"%s\" cannot be used as a path name\n",name));
	return RET_FAIL;
    }
    for (s=path; *s && *s>' '; s++);
    if (*s) {
	pr("Path string cannot contain blank characters\n");
	return RET_FAIL;
    }
    sprintf(buf,"%s%d",pathnamefil,cn);
    if ((fout=fopen(buf,"a"))) {
	fprintf(fout,"%s %s\n",name,path);
	fclose(fout);
    }
    return RET_OK;
}

shownamedpaths(cn)
natid cn;
{
    FILE *fin;
    char *s,*e;
    char buf[1024];
    sprintf(buf,"%s%d",pathnamefil,cn);
    if ((fin=fopen(buf,"r"))) {
	while (fgets(buf,1024,fin)) {
	    for (s=buf; *s>' '; s++); e=s;
	    while (*s && *s<=' ') s++;
	    *e = '\0';
	    pr(fmt("  %10s  %s\n",buf,s));
	}
	fclose(fin);
    }
    return RET_OK;
}

char *getnamedpath(name,namelen,cn)
char *name;
int namelen;
natid cn;
{
    FILE *fin;
    char *s,*t,*u;
    char buf[1024];
    t = (char *) 0;
    sprintf(buf,"%s%d",pathnamefil,cn);
    if ((fin=fopen(buf,"r"))) {
	while (fgets(buf,1024,fin)) {
	    if (!strncmp(name,buf,namelen)) {
		s = buf+namelen;
		if (*s && *s<=' ') {
		    while (*s && *s<=' ') s++;
		    if (*s) {
			t = u = (char *) malloc(strlen(s));
			while (*s && *s>' ') *u++ = *s++;
			*u = '\0';
			break;
		    }
		}
	    }
	}
	fclose(fin);
    }
    return t;
}

