/*
**  spysub.c
**
**  Miscellaneous spy functions
**
**  7/24/91 Bill Frolik
*/

#include "misc.h"
#include "var.h"
#include "spy.h"
#include "xy.h"
#include "file.h"
#include "nat.h"
#include "deity.h"
#include "ship.h"
#include "sect.h"
#include "nsc.h"
#include "plane.h"
#include "item.h"
#include "nuke.h"
#include "news.h"
#include "path.h"
#include "loc.h"

/*
**  NOTE: Many of these routines depend on the initial elements
**  of (struct sctstr) and (struct shpstr) being identical -- in
**  particular the owner, x- and y-locations, and efficiency
**  fields.
*/

/* ________________________________________________________________
**|                                                                |
**| spyleave - make a spy leave a ship or vehicle                  |
**|________________________________________________________________|
*/

spyleave(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    struct sctstr sect;
    struct shpstr ship;
    coord x,y;
    int dir;
    if (spy->spy_ship<0) return 0;
    dpr(fmt("spyleave: %s",spyas(spy,LOPS)));
    if (!getship(spy->spy_ship,&ship)) return 0;
    dpr(fmt("spyleave: got %s",shipas(&ship,LOPS)));
    if (ship.shp_depth>0) {
	pr(fmt("Can't leave %s while %s.\n", shipas(&ship,0),
	       isballoon(&ship) ? "airborne" : "submerged"));
    }
    if (!spyhasmobil(spy)) return 1;
    if (cantdoit(spy,&ship,EF_SHIP,diff,"attempting to leave")) return 1;
    pr(fmt("%s ",spyas(spy,0)));
    if (ship.shp_aboard >= 0) {
	spy->spy_ship = ship.shp_aboard;
	pr(fmt("leaves %s, ",shipas(&ship,LO)));
	getship(ship.shp_aboard,&ship);
	pr(fmt("now %s\n",shipas(&ship,LOP)));
    } else {
	getsect(ship.shp_x,ship.shp_y,&sect);
	if (isship(&ship) && sect.sct_type==SCT_WATER)
	    if (!argp[3] || !*argp[3]) {
		pr("can't leave at sea without a direction\n");
	    } else if ((dir=chkdir(*argp[3],DIR_FIRST,DIR_LAST)) == -1) {
		pr(fmt("\"%c\" is not a valid direction", *argp[3]));
	    } else {
		x = sect.sct_x + diroff[dir][0];
		y = sect.sct_y + diroff[dir][1];
		getsect(x,y,&sect);
		if (sect.sct_type==SCT_WATER)
		    pr("There's nothing out there but water!\n");
		else if (sect.sct_type < SCT_RURAL)
		    pr("Can't go ashore there.\n");
		else {
		    pr(">splash< slips ashore in a rubber raft.\n");
		    spy->spy_ship = -1;
		    spy->spy_x = sect.sct_x;
		    spy->spy_y = sect.sct_y;
		    spy->spy_mobil--;
		    cantdoit(spy,&sect,EF_SECTOR,30,"coming ashore");
		}
	    }
	else {
	    pr(fmt("is now on foot %s\n",sectas(&sect,LOP)));
	    spy->spy_ship = -1;
	    spy->spy_x = sect.sct_x;
	    spy->spy_y = sect.sct_y;
	}
    }
    spy->spy_mobil--;
    putspy(spy->spy_uid,spy);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spycoup - try to take a sector                                 |
**|________________________________________________________________|
*/

spycoup(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    
}

/* ________________________________________________________________
**|                                                                |
**| spyretire - return to military life                            |
**|________________________________________________________________|
*/

spyretire(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int mil,civ,vested,pension;
    struct sctstr *sp;
    natid own;
    extern int etu_per_update;
    extern int spy_retire;
    extern double retire_pay;
    own = ((struct sctstr *)loc)->sct_own;
    if (own && own!=cnum)
	pr(fmt("%s can't retire now!\n",spyas(spy,0)));
    else {
	mil = getvar(V_MILIT,loc,loctype);
	civ = getvar(V_CIVIL,loc,loctype);
	putvar(V_MILIT,++mil,loc,loctype);
	vested = etu_per_update * spy->spy_age * 10 / spy_retire * 10;
	if (vested > 40) { /* vested is 50, 60, 70, 80, 90, or 100 */
	    pension = vested * retire_pay / 100;
	    pr(fmt("%s had %d%% vested pension, paid $%d\n",vested,pension));
	    dolcost += pension;
	}
	pr(fmt("%s resumes normal military duties %s\n",
	       spyas(spy,0),locas(loc,loctype,LPS)));
	spy->spy_own = 0;
	spy->spy_effic = 0;
	putspy(spy->spy_uid,spy);
	if (!own || (!mil && !civ)) {
	    sp = (struct sctstr *)loc;
	    sp->sct_own = cnum;
	    if (loctype==EF_SECTOR) {
		takeover(sp,0);
		sp->sct_oldown = cnum;
		sp->sct_loyal = 0;
		sp->sct_work = 100;
	    }
	    putplace(loc,loctype);
	}
    }
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyset - DEITY COMMAND - change a spy's efficiency/mobility    |
**|________________________________________________________________|
*/

spyset(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i;
    char *p,*d,*t;
    p = argp[3];
    switch (*argp[2]) {
    case 'm': 
	t = "mob";
	d = &(spy->spy_mobil);
	break;
    case 'e':
	t = "eff";
	d = &(spy->spy_effic);
	break;
    default:
	return 0;
    }
    if (!p || !isdigit(*p)) 
	p = getstarg(p,fmt("%s : %s (%d) ? ", spyas(spy,LOS), t, *d));
    if (!p || !isdigit(*p)) 
	return 1;
    i = atoi(p);
    if (i<0) i=0;
    if (*t == 'e') {
	if (i>100) i=100;
    } else {
	if (i>127) i=127;
    }
    *d = i;
    pr(fmt("%s",spyas(spy,LOS)));
    putspy(spy->spy_uid,spy);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spygive - give a spy to another country                        |
**|________________________________________________________________|
*/

spygive(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    natid newown;
    struct natstr *natp;
    newown = natarg(argp[3], fmt("%s: Give to whom? ",spyas(spy,LS)));
    if (newown < 0) return 0;
    if ((natp=getnatp(newown)) == 0) return 0;
    if (natp->nat_stat & STAT_INUSE) {
	pr(fmt("%s given to %s\n",spyas(spy,0),cname(newown)));
	wu(0, newown, 
	   fmt("%s has given you %s", cname(spy->spy_own), spyas(spy,0)));
	if (spy->spy_effic > 25) spy->spy_effic -= 20;
	if (spy->spy_effic < 25) spy->spy_effic = 25;
	if (spy->spy_mobil > 0)	 spy->spy_mobil -= 20;
	if (spy->spy_mobil < 0)	 spy->spy_mobil = 0;
	spy->spy_own = newown;
	putspy(spy->spy_uid,spy);
	return 1;
    }
    return 0;
}

/* ________________________________________________________________
**|                                                                |
**| spyboard - make a spy board a ship or vehicle                  |
**|________________________________________________________________|
*/

spyboard(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i;
    char *p;

    struct nstr_item ni;
    struct shpstr ship;
    p = getstarg(argp[3],fmt("%s: Board what? ",spyas(spy,LS)));

    if (!p || !*p) {
	dpr("spyboard: null response");
	return 0;
    }

    if (!snxtitem(&ni, EF_SHIP, p)) {
	pr(fmt("Ship/vehicle '%s' doesn't exist.\n",p));
	return 0;
    }

    ni.flags &= ~EFF_OWNER;
    if (!nxtitem(&ni, (char *)&ship)) {
	logerror(fmt("spyboard: ship #%d nxtitem failure",ni.cur));
	return 0;
    }

    if (ship.shp_x!=spy->spy_x || ship.shp_y!=spy->spy_y ||
	(issub(&ship) && ship.shp_depth>0)) {
	pr("Don't see that here.\n");
	return 1;
    }

    if (ship.shp_depth>0) {
	pr(fmt("Can't board %s while it's airborne.\n", shipas(&ship,0)));
	return 1;
    }

    if (!spyhasmobil(spy))
	return 1;

    if (cantdoit(spy,&ship,EF_SHIP,diff,"attempting to board")) 
	return 1;

    pr(fmt("%s ",spyas(spy,0)));
    if (ship.shp_aboard!=spy->spy_ship) {
	pr("Can't do that here\n");
	return 1;
    }

    pr(fmt("now %s\n", shipas(&ship,LOP)));
    spy->spy_ship = ship.shp_uid;
    spy->spy_mobil--;
    putspy(spy->spy_uid,spy);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spybomb - bomb something                                       |
**|								   |
**| Pinpoint bomb a specific commodity.  Volatile targets (oil,    |
**| can lead to chain reaction explosions.                         |
**|________________________________________________________________|
*/

spybomb(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc;
int diff; /* unused */
{
    char *p;
    natid victim;
    int i,j,chain,err,shells,iitem,item,nitem;
    int triggers,score,peopletoo;
    int	vec[I_MAX+1];
    double coll;
    struct ichrstr *ip;
    struct sctstr *sect;
    struct shpstr *ship;
    if (!qualifiedspy(spy,S_B,0)) return 0;
    if (!spyhasmobil(spy)) return 1;
    shells = getvar(V_SHELL, (char *)spy, EF_SPY);
    if (!shells)
	pr(fmt("%s: needs a shell\n",spyas(spy,LS)));
    else {
	spyreport(spy,argp,loc,loctype,0);
	err = errormargin(spy);
	if (!(ip = whatitem(argp[3],"!Commodity to bomb? ")))
	    return 1;
	item = ip->i_vtype;
	if ((iitem=nitem=getvar(item,loc,loctype)) == 0) {
	    pr(fmt("No %s here\n",ip->i_name));
	    return 1;
	}
	if (item==V_CIVIL || item==V_MILIT) {
	    if (!qualifiedspy(spy,S_A,0)) {
		pr(fmt("%s: refuses to bomb %s\n",
		       spyas(spy,LS),ip->i_name));
		return 1;
	    }
	}
	if (cantdoit(spy,loc,loctype,ip->i_secure,"planting bomb")) return 1;
	pr("tick...  tick...  tick...\n");
	putvar(V_SHELL,--shells,(char *)spy,EF_SPY);
	triggers = 1;
	chain = 0;
	coll = 100.0;
    chainreaction:
	for (i=0; i<min(triggers,5); i++) {
	    coll *= (loctype==EF_SHIP ? 0.92 : 0.99);
	    switch (roll(4)) {
	    case 1: pr(" K A B O O M !!!"); break;
	    case 2: pr(" K A P O W !!!"); break;
	    case 3: pr(" BLAM!!"); break;
	    case 4: pr(" FWOOOM!"); break;
	    };
	}
	pr("\n");
	if (err > 100) err = 100;
	score = roll(5+triggers) + (roll(80)*(100-err)/100) - 1;
	if (score > nitem) score = nitem;
	nitem -= score;
/*	pr(fmt("%d %s %sed in %sblast\n",
	       score, ip->i_name, 
	       (item==V_CIVIL || item==V_MILIT || item==V_UW)
	       ? "kill" : "destroy", chain ? "chain reaction " : "")); */
	if (item==V_PETROL || item==V_SHELL || item==V_OIL) {
	    chain++;
	    err += ((loctype==EF_SHIP) ? roll(5)+2 : roll(10)+5);
	    triggers = min(roll(score+1)-1,10);
	    if (triggers > ((loctype==EF_SHIP)?0:3))
		goto chainreaction;
	}
	score = iitem - nitem;
	coll = 100.0 - coll;
	pr(fmt("Total damage: %d %s, %.1f%% collateral\n",
	       score, ip->i_name, coll));
	putvar(item,nitem,loc,loctype);
	spy->spy_mobil -= 20;
	putspy(spy->spy_uid,spy);
	if (loctype == EF_SHIP) {
	    ship = (struct shpstr *)loc;
	    victim = ship->shp_own;
	    wu(0, victim, 
	       fmt("Explosions %sed %d %s and did %.1f%% damage on %s #%d!",
		   (item==V_CIVIL || item==V_MILIT || item==V_UW)
		   ? "kill" : "destroy",
		   score, ip->i_name, coll,
		   mchr[ship->shp_type].m_name, ship->shp_uid));
	    shipdamage(ship,(int)coll);
	} else {
	    sect = (struct sctstr *)loc;
	    victim = sect->sct_own;
	    wu(0, victim,
	       fmt("Explosions %sed %d %s and did %.1f%% damage in %s!",
		   (item==V_CIVIL || item==V_MILIT || item==V_UW)
		   ? "kill" : "destroy",
		   score, ip->i_name, coll,
		   xyas(spy->spy_x,spy->spy_y,victim)));
	    sectdamage(sect, (int)coll);
	}
	putplace(loc,loctype);
    }
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spysteal - steal a plane                                       |
**|________________________________________________________________|
*/

spysteal(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc;
int diff;
{
    int i;
    natid oldown;
    char *p,*id,*s;
    struct plnstr plane;
    char pidbuf[16],xybuf[16];
    double odds;
    
    spyreport(spy,argp,loc,loctype,0);
    p = getstarg(argp[3],fmt("%s: Plane id? ",spyas(spy,LS)));
    if (!p || !*p) return 0;
    if (!getplane(atoi(p),&plane)) {
	pr("Bad plane number\n");
	return 1;
    }
    if (plane.pln_ship >= 0) 
	i = (plane.pln_ship == spy->spy_ship);
    else 
	i = ((plane.pln_x == spy->spy_x) && (plane.pln_y == spy->spy_y));
    if (!plane.pln_own || !i) {
	pr("Don't see that aircraft here.\n");
	return 1;
    }

    if (plchr[plane.pln_type].pl_flags & (P_M|P_O)) {
	pr("You can't fly that kind of aircraft.\n");
	return 1;
    }

    if (plane.pln_own == spy->spy_own) {
	pr("You can't steal your own aircraft.\n");
	return 1;
    }

    if (((struct sctstr *)loc)->sct_own == cnum) {
	pr("You can't steal a plane from your own ");
	if (loctype==EF_SHIP) pr("ship.\n");
	else pr("sector.\n");
	return 1;
    }

    if (plane.pln_flags & PLN_BOXED) {
	pr(fmt("%d is disassembled and crated\n",plane.pln_uid));
	return 1;
    }

    if (!pln_oktotakeoff(&plane))
	return 1;

    if (plchr[plane.pln_type].pl_fuel > getvar(V_PETROL,loc,loctype)) {
	pr(fmt("%s not enough petrol there!\n",planeas(&plane,0)));
	return 1;
    }

    odds = (double)spy->spy_plntech/(double)plane.pln_tech;
    if (odds > 1.0) odds = 1.0;
    if (odds < 1.0) {
	pr(fmt("Spy has %.1f%% of training needed to fly %s.\n",
	       odds*100.0,planeas(&plane,0)));
	s = getstring("Go for it anyway? ");
	if (!s || !(*s=='y' || *s=='Y')) return 1;
    }

    if (cantdoit(spy,loc,loctype,diff,"boarding aircraft")) return 1;

    oldown = plane.pln_own;
    plane.pln_own = cnum;

    if (chance(1.0-odds)) {
	pr("  CRASH!!! -- Nice takeoff, Lindbergh.\n");
	wu(0, oldown, 
	   fmt("%s crashed after unauthorized takeoff in %s.",
	       planeas(&plane,0),locasc(loc,loctype,LS,oldown)));
	plane.pln_own = 0;
	plane.pln_effic = 0;
	putplane(plane.pln_uid,&plane);
	spy->spy_effic = 0;
	spy->spy_own = 0;
    } else {
	plane.pln_mobil = 20; /* enough for one flight */
	putplane(plane.pln_uid,&plane);
	wu(0, oldown, 
	   fmt("%s stolen from %s.",
	       planeas(&plane,0),locasc(loc,loctype,LS,oldown)));
	strncpy(pidbuf,p,9); pidbuf[9]='\0';
	sprintf(xybuf,"%s",xyas(spy->spy_x, spy->spy_y, cnum));
	argp[1] = pidbuf;
	argp[3] = xybuf;
	argp[2] = argp[4] = "";
	argp[5] = ".";
	fly();
	if (getplane(plane.pln_uid,&plane) && plane.pln_own) {
	    if (spy->spy_x==plane.pln_x && 
		spy->spy_y==plane.pln_y && 
		spy->spy_ship==plane.pln_ship) {
		if (getvar(V_MILIT,loc,loctype)) {
		    pr(fmt("%s captured by local military.\n",spyas(spy,0)));
		    wu(0, oldown,
		       fmt("%s recovered; %s captured.",
			   planeas(&plane,0),spyas(spy,LO)));
		    nreport(oldown,N_CAPTURE_SPY,cnum,1);
		    plane.pln_own = oldown;
		    spy->spy_own = oldown;
		    spy->spy_effic = 10;
		    spy->spy_mobil = 0;
		    spy->spy_hap = 0;
		}
	    } else {
		spy->spy_x = plane.pln_x;
		spy->spy_y = plane.pln_y;
		spy->spy_ship = plane.pln_ship;
	    }
	} else {
	    spy->spy_effic = 0;
	    spy->spy_own = 0;
	}
    }
    putspy(spy->spy_uid,spy);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spytake - pick up something                                    |
**|________________________________________________________________|
*/

spytake(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc;
int diff; /* unused */
{
    register struct ichrstr *ich;
    register int item;
    natid host;
    int sect_amt,spy_amt;
    spyreport(spy,argp,loc,loctype,0);
    if (!(ich=whatitem(argp[3],fmt("!%s: Take what? ",spyas(spy,LS))))) {
	return 0;
    }
    item = ich->i_vtype;
    if (item==V_MILIT || item==V_CIVIL || item==V_UW) {
	pr(fmt("%s can't carry people!\n",spyas(spy,0)));
	return 0;
    }
    spy_amt = getvar(item, (char *)spy, EF_SPY);
    if (spy_amt) {
	pr(fmt("%s can't carry any more %s.\n",spyas(spy,0),ich->i_nam)); 
	return 0;
    }
    sect_amt = getvar(item, loc, loctype);
    if (!sect_amt) {
	pr(fmt("No %s here!\n",ich->i_name)); 
	return 0;
    }
    if (cantdoit(spy,loc,loctype,ich->i_secure,"taking item")) return 1;
    if (putvar(item, ++spy_amt, (char *)spy, EF_SPY)) {
	putvar(item, --sect_amt, loc, loctype);
	host = (natid)placeown(loc,loctype);
	/* Item's old owner might discover that something's missing */
	if (host != cnum) {
	    sect_amt = getvar(V_MILIT,loc,loctype)*2
		+ getvar(V_CIVIL,loc,loctype);
	    if (loctype==EF_SHIP) sect_amt *= 5;
	    if (chance((double)sect_amt/1500.0)) {
		wu(0, host,
		   fmt("%s stolen %s",
		       ich->i_name, locasc(loc,loctype,LPS,host)));
	    }
	}
    }
    putspy(spy->spy_uid,spy);
    putplace(loc,loctype);
    pr(fmt("%s taken\n",ich->i_nam));
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spydrop - drop something                                       |
**|________________________________________________________________|
*/

spydrop(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    register struct ichrstr *ich;
    register int item;
    int sect_amt,spy_amt;
    spyreport(spy,argp,loc,loctype,0);
    if (!(ich=whatitem(argp[3],fmt("%s: Drop what? ",spyas(spy,LS))))) {
	pr("Nothing dropped\n");
	return 0;
    }
    item = ich->i_vtype;
    spy_amt = getvar(item, (char *)spy, EF_SPY);
    if (!spy_amt) {
	pr(fmt("%s isn't carrying any %s!\n",spyas(spy,0),ich->i_nam)); 
	return 0;
    }
    if (cantdoit(spy,loc,loctype,diff,0)) return 1;
    sect_amt = getvar(item, loc, loctype);
    if (putvar(item, ++sect_amt, loc, loctype))
	putvar(item, --spy_amt, (char *)spy, EF_SPY);
    putspy(spy->spy_uid,spy);
    putplace(loc,loctype);
    pr(fmt("%s dropped\n",ich->i_nam));
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyreport - report spy's own location                          |
**|________________________________________________________________|
*/

spyreport(spy,argp,loc,loctype,vc)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc;
{
    struct shpstr *ship;
    struct sctstr sct,*sect;
    struct spchrstr *spc;
    int vec[I_MAX+1],i,j,err;
    spc = &spchr[spy->spy_type];
    pr(fmt("%s (%d%%) ", spyas(spy,0), spy->spy_effic));
    err = errormargin(spy);
    if (spy->spy_ship >= 0) {
	ship = (struct shpstr *)loc;
	getsect(ship->shp_x,ship->shp_y,&sct);
	sect = &sct;
	pr(fmt("%s",shipas(ship,LOPS)));
    } else {
	sect = (struct sctstr *)loc;
	pr(fmt("%s",sectas(sect,LOPS)));
    }
    setbigmap(sect->sct_x, sect->sct_y, dchr[sect->sct_type].d_mnem);
    getvec(VT_ITEM, vec, (char *)spy, EF_SPY);
    for (j=0,i=1; i<I_MAX; i++) if (vec[i]) {
	if (!j++) pr(" [");
	if ( j>1) pr(" ");
	pr(fmt("%c:%d",ichr[i].i_mnem,vec[i]));
    }
    if (j) pr("]");
    pr("\n");
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyreportlev - report distribution levels                      |
**|________________________________________________________________|
*/

spyreportlev(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int err;
    if (!qualifiedspy(spy,S_I,0)) return 0;
    if (loctype==EF_SHIP && spy->spy_ship<0) return 0;
    if (loctype==EF_SECTOR && spy->spy_ship>=0) return 0;
    spyreport(spy,argp,loc,loctype,0);
    if (cantdoit(spy,loc,loctype,diff,"acquiring info")) return 1;
    err = errormargin(spy);
    showlevels(loc,0,1,err);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyreportcom - report commodity levels                         |
**|________________________________________________________________|
*/

spyreportcom(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int err;
    if (!qualifiedspy(spy,S_I,0)) return 0;
    if (loctype==EF_SHIP && spy->spy_ship<0) return 0;
    if (loctype==EF_SECTOR && spy->spy_ship>=0) return 0;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    if (cantdoit(spy,loc,loctype,diff,"acquiring info")) return 1;
    if (loctype==EF_SHIP)
	showcargo(loc,1,err);
    else
	showcommodities(loc,1,err);
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyreportcen - report census levels                            |
**|________________________________________________________________|
*/

spyreportcen(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int err;
    if (loctype==EF_SHIP && spy->spy_ship<0) return 0;
    if (loctype==EF_SECTOR && spy->spy_ship>=0) return 0;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    if (loctype==EF_SHIP) {
	if (cantdoit(spy,loc,loctype,diff*4,"acquiring info")) return 1;
	showvehicle(loc,1,err);
    } else {
	if (cantdoit(spy,loc,loctype,diff,"acquiring info")) return 1;
	showcensus(loc,1,err);
    }
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyrecon - recon the surrounding sectors                       |
**|________________________________________________________________|
*/

spyrecon(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i,nsects,err,vec[I_MAX+1];
    coord x,y;
    struct sctstr sect;
    struct spchrstr *scp;
    if (spy->spy_ship >= 0) return 0;
    if (!getsect(spy->spy_x,spy->spy_y,&sect)) return 0;
    if (!spyhasmobil(spy)) return 1;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    scp = &spchr[spy->spy_type];
    for (nsects=i=0; i<=6; i++) {
	x = spy->spy_x + diroff[i][0];
	y = spy->spy_y + diroff[i][1];
	if (!getsect(x, y, &sect)) continue;
	setbigmap(x,y,dchr[sect.sct_type].d_mnem);
	if (sect.sct_type==SCT_WATER || sect.sct_type==SCT_WASTE) continue;
	if (cantdoit(spy,&sect,EF_SECTOR,diff,0)) {
	    if (spy->spy_own != cnum) return 1;
	    pr(fmt("  can't recon %s\n",xyas(sect.sct_x,sect.sct_y,cnum)));
	} else {
	    spy->spy_mobil--;
	    showneighbor(&sect,(nsects++ == 0),err);
	}
    }
    putspy(spy->spy_uid,spy);
    return 1;
}
    
showneighbor(sp,header,plusorminus)
struct sctstr *sp;
int header,plusorminus;
{
    int vec[I_MAX+1];
    if (header) {
	prest(plusorminus);
	pr("   sect   type own  eff  civ  mil  shl  gun  iron  pet  food\n");
    }
    getvec(VT_ITEM, vec, (char *) sp, EF_SECTOR);
    pr(xyfmt("%4d,%-4d",sp->sct_x, sp->sct_y, cnum));
    pr(fmt("   %c %4d %4d %4d %4d %4d %4d  %4d %4d %5d\n",
	   dchr[sp->sct_type].d_mnem, sp->sct_own,
	   limestimate((int)sp->sct_effic,plusorminus,0,100),
	   estimate(vec[I_CIVIL],plusorminus),
	   estimate(vec[I_MILIT],plusorminus),
	   estimate(vec[I_SHELL],plusorminus), 
	   estimate(vec[I_GUN],plusorminus),
	   estimate(vec[I_IRON],plusorminus), 
	   estimate(vec[I_PETROL],plusorminus),
	   estimate(vec[I_FOOD],plusorminus)
	   ));
}

spylook(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i,j,dx,dy,nsects;
    char map[3][6];
    struct sctstr sect;
    coord x,y;
    if (!getsect(spy->spy_x,spy->spy_y,&sect)) return 0;
    spyreport(spy,argp,loc,loctype,0);
    for (i=0; i<3; i++)	strcpy(map[i],"     ");
    for (i=0; i<=6; i++) {
	dx = diroff[i][0];
	dy = diroff[i][1];
	x = spy->spy_x + dx;
	y = spy->spy_y + dy;
	if (!getsect(x, y, &sect)) continue;
	setbigmap(x,y,dchr[sect.sct_type].d_mnem);
	map[dy+1][dx+2] = dchr[sect.sct_type].d_mnem;
    }
    for (i=0; i<3; i++)
	pr(fmt("   %s\n",map[i]));
    return 1;
}
    
/* ________________________________________________________________
**|                                                                |
**| spyreportplanes - report planes in this sector                 |
**|________________________________________________________________|
*/

spyreportplanes(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i,err;
    struct nstr_item ni;
    struct plnstr plane;

    if (!snxtitem(&ni, EF_PLANE,"*")) return 0;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    if (cantdoit(spy,loc,loctype,diff,"acquiring info")) return 1;

    for (i=0; nxtitem(&ni, (char *)&plane);) {
	if (!plane.pln_own 
	    || plane.pln_ship!=spy->spy_ship 
	    || plane.pln_x!=spy->spy_x 
	    || plane.pln_y!=spy->spy_y )
	    continue;
	showplane(&plane,!i++,err);
    }

    if (!i) pr("No planes\n");
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spyreportvc - report other vessels/vehicles in this sector     |
**|________________________________________________________________|
*/

spyreportvc(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    int i,j,k,err,nships,totalships;
    natid cn;
    struct nstr_item ni;
    struct shpstr ship;
    struct ss {
	natid owner;
	int snum;
    } *shiplist;

    if (!snxtitem(&ni, EF_SHIP,"*")) return 0;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    if (cantdoit(spy,loc,loctype,diff,"acquiring info")) return 1;

    shiplist = (struct ss *) calloc(1024,sizeof(struct ss));
    for (nships=0; nxtitem(&ni, (char *)&ship);) {
	if (   ship.shp_type<0
	    || ship.shp_type>shp_maxno
	    || !ship.shp_own
	    || mchr[ship.shp_type].m_class != vc
	    || ship.shp_x!=spy->spy_x
	    || ship.shp_y!=spy->spy_y
	    || (ship.shp_depth>0 && ship.shp_uid!=spy->spy_ship)
	    || ship.shp_aboard!=spy->spy_ship )
	    continue;
	shiplist[nships].owner = ship.shp_own;
	shiplist[nships].snum  = ship.shp_uid;
	nships++;
    }

    totalships = 0;
    for (cn=1; cn<MAXNOC; cn++) {
	k = 0;
	for (j=0; j<nships; j++)
	    if (shiplist[j].owner == cn) {
		getship(shiplist[j].snum,&ship);
		if (!k++) {
		    i = 0;
		    pr(fmt("\n%s (#%d) %ss:\n",cname(cn),cn,class[vc].name));
		}
		if (vcshowvehicle(&ship,(i==0),err,vc)) {
		    i++; totalships++;
		}
	    }
    }

    free(shiplist);
    if (!totalships) pr(fmt("No %ss\n",class[vc].name));
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spykill - attempt to kill a spy, civ, mil, or uw               |
**|________________________________________________________________|
*/

spykill(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    char *p;
    int i,j,err,qty,vec[I_MAX+1];
    struct nstr_item ni;
    int fail,targid,targtype;
    struct spystr targspy;
    struct sctstr *sect;
    struct shpstr *ship;
    if (!qualifiedspy(spy,S_A,0)) return 0;
    if (!spyhasmobil(spy)) return 1;
    if (loctype==EF_SECTOR) sect = (struct sctstr *)loc;
    else ship = (struct shpstr *)loc;
    err = errormargin(spy);
    spyreport(spy,argp,loc,loctype,0);
    p = getstarg(argp[3],"Kill what (civ, mil, uw, spy#)? ");
    switch (*p) {
    case 'c': targtype = V_CIVIL; break;
    case 'm': targtype = V_MILIT; break;
    case 'u': targtype = V_UW; break;
    default:  
	targtype = 0;
	if (!isdigit(*p)) {
	    pr("Target must be a spy id, or 'civ', 'mil', or 'uw'\n"); 
	    return 0;
	}
    }
    if (targtype > 0) {
	if (cantdoit(spy,loc,loctype,diff,0)) return 1;
	if ((qty=getvar(targtype, loc, loctype)) > 0) {
	    qty--;
	    putvar(targtype,qty,loc,loctype);
	    pr(fmt("  >bang!< one %s liquidated %s\n",
		   ichr[unitem(targtype)].i_name,locas(loc,loctype,LPOS)));
	    if (loctype==EF_SECTOR) {
		putsect(sect);
		if (qty>0 && chance(0.5))
		    wu(0,sect->sct_own,fmt("1 %s found dead %s",
					   ichr[unitem(targtype)].i_name,
					   sectasc(sect,LPS,sect->sct_own)));
	    } else {
		putship(ship->shp_uid,ship);
		if (qty>0 && chance(0.5))
		    wu(0,ship->shp_own,fmt("1 %s found dead %s",
					   ichr[unitem(targtype)].i_name,
					   shipasc(ship,LPS,ship->shp_own)));
	    }
	    spy->spy_mobil -= 5;
	    putspy(spy->spy_uid,spy);
	}
    } else {
	if (getspy(atoi(p),&targspy)) {
	    if (targspy.spy_x==spy->spy_x && targspy.spy_y==spy->spy_y &&
		targspy.spy_ship==spy->spy_ship) {
		if (targspy.spy_uid == spy->spy_uid) 
		    pr("  Suicide?  Well, okay ...\n");
		else if (cantdoit(spy,loc,loctype,diff*3,0))
		    return 1;
		if (chance((double)err/100.0)) {
		    pr(fmt("  >bang!< ... missed\n"));
		    wu(0, targspy.spy_own, 
		       fmt("Sniper shot at %s %s",
			   spyas(&targspy,0),locas(loc,loctype,LPOS)));
		    targspy.spy_effic -= (roll(5)-1);
		} else {
		    pr(fmt("  >bang!< %s terminated\n", spyas(&targspy,LO)));
		    wu(0, targspy.spy_own, 
		       fmt("Sniper killed %s %s",
			   spyas(&targspy,0),locas(loc,loctype,LPOS)));
		    targspy.spy_effic = 0;
		}
		putspy(targspy.spy_uid,&targspy);
	    }
	    spy->spy_mobil -= 10;
	    putspy(spy->spy_uid,spy);
	}
    }
    return 1;
}

/* ________________________________________________________________
**|                                                                |
**| spymove - move a spy                                           |
**|________________________________________________________________|
*/

spymove(spy,argp,loc,loctype,vc,diff)
struct spystr *spy;
char *argp[],*loc;
int loctype,vc,diff;
{
    coord newx,newy;
    struct sctstr sect;
    struct shpstr ship;
    char *p,buf[80];
    double mobcost;
    int err,i,dir;

    p = argp[3];
    err = errormargin(spy);
    pr(fmt("%s\n",spyas(spy,LS)));
    if (spy->spy_ship >= 0) {
	pr("  must be on solid ground to move\n");
	return 1;
    }
    if (!spyhasmobil(spy)) return 1;
    while (spy->spy_effic>=5 && spy->spy_mobil>0) {
	if (!p || !*p) {
	    pr("");
	    getsect(spy->spy_x, spy->spy_y, &sect);
	    sprintf(buf,"<%.1f:%c:%s> ",
		    (double)spy->spy_mobil,
		    (char)dchr[sect.sct_type].d_mnem,
		    (char *)xyas(spy->spy_x,spy->spy_y,cnum));
	    spylook(spy,argp,&sect,EF_SECTOR,0,0);
	    p = getstring(buf);
	}
	if (!p || !*p) break;
	if ((dir=chkdir(*p++, DIR_STOP, DIR_VIEW)) == -1) {
	    pr(fmt("\"%c\" is not legal...", p[-1]));
	    direrr("`%c' to stop\n`r' to recon neighbors",0,0);
	    p=0; continue;
	}
	if (dir == DIR_STOP) break;
	if (dir == DIR_VIEW) {
	    if (sect.sct_type==SCT_WATER)
		pr("at sea\n");
	    else
		pr(fmt("%d%% efficient %s %s\n",
		       limestimate((int)sect.sct_effic,err,0,100),
		       cname(sect.sct_own),dchr[sect.sct_type].d_name));
	    continue;
	}
	mobcost = spy->spy_effic * 0.30;
	mobcost = 100.0 / (mobcost + techfact(spy->spy_tech, mobcost));
	if (mobcost > 64.0) mobcost = 64.0;
	/*
	**  Exit the current sector
	*/
	if (sect.sct_own != spy->spy_own) {
	    if (cantdoit(spy,&sect,EF_SECTOR,diff,
			 fmt("exiting %s", xyas(sect.sct_x,sect.sct_y,cnum)))) {
		if (spy->spy_own != cnum) return 1;
		pr(fmt("%s failed to leave %s\n",
		       spyas(spy,0), xyas(spy->spy_x,spy->spy_y,cnum)));
		if (!god) spy->spy_mobil -= mobcost/2;
		break;
	    } 
	}
	/*
	**  Enter the new sector
	*/
	newx = xnorm(spy->spy_x + diroff[dir][0]);
	newy = ynorm(spy->spy_y + diroff[dir][1]);
	if (!getsect(newx, newy, &sect)) {
	    logerror("Sector failure");
	    p=0; continue;
	}
	mobcost *= dchr[sect.sct_type].d_mcst;
	if (mobcost > 64.0) mobcost = 64.0;
	switch (sect.sct_type) {
	case SCT_WATER:
	case SCT_WASTE:
	case SCT_SANCT:
	    pr("Can't go there\n");
	    p=0; continue;
	}
	if (sect.sct_own != spy->spy_own) {
	    if (cantdoit(spy,&sect,EF_SECTOR,diff,
			 fmt("entering %s",sectas(&sect,LOS)))) {
		if (spy->spy_own != cnum) return 1;
		pr(fmt("%s failed to enter %s\n",
		       spyas(spy,0), xyas(newx,newy,cnum)));
		if (!god) spy->spy_mobil -= mobcost/2;
		break;
	    } 
	}
	spy->spy_x = newx;
	spy->spy_y = newy;
	if (!god) spy->spy_mobil -= mobcost;
    }
    if (spy->spy_effic>=5) {
	getsect(spy->spy_x,spy->spy_y,&sect);
	if (spy->spy_mobil>0.0)
	    pr(fmt("%s stopped at %s\n",
		   spyas(spy,0),sectas(&sect,LOS)));
	else
	    pr(fmt("%s out of mobility at %s\n",
		   spyas(spy,0),sectas(&sect,LOS)));
    }
    putspy(spy->spy_uid,spy);
    return 1;
}

/* ______________________________________________________________________
**|                                                                      |
**| qualifiedspy - return TRUE if spy has ALL of the "want"              |
**|       requirements and NONE of the "dontwant" requirements.          |
**|______________________________________________________________________|
*/

qualifiedspy(spy,want,dontwant)
struct spystr *spy;
int want,dontwant;
{
    int goodman;
    goodman = !(spchr[spy->spy_type].sp_flags & dontwant);
    if (goodman) goodman = ((spchr[spy->spy_type].sp_flags & want)==want);
    if (!goodman)
	pr(fmt("%s doesn't do that kind of work\n",spyas(spy,0)));
    return (goodman);
}


spyhasmobil(spy)
struct spystr *spy;
{
    if (spy->spy_mobil <= 0) {
	pr(fmt("%s has no mobility\n",spyas(spy,0)));
	return 0;
    }
    return 1;
}

/* ______________________________________________________________________
**|                                                                      |
**| cantdoit - Given spy, mission locale, and difficulty factor,         |
**|       if the spy bungles the mission, return the id of the country   |
**|       that caused him to mess up, or zero if he does the job without |
**|       any problem.  Difficulty range is 0 (trivial) to 100 (almost   |
**|       impossible).                                                   |
**|______________________________________________________________________|
*/

cantdoit(spy,loc,loctype,difficulty,riskmsg)
struct spystr *spy;
char *loc,*riskmsg;
int loctype,difficulty;
{
    coord x,y;
    char *note;
    natid milown;
    struct natstr nat,*natp;
    struct shpstr *ship;
    struct sctstr *sect;
    struct spchrstr *spc;
    int i,mil,detected,dam,relat;
    double mileff,miledu,miltech;
    double spyeff,spyedu,spytech;
    double detect,spygoof,risk,capture,d;

    spyeff = (double)spy->spy_effic/100.0;
    spyedu = (double)spy->spy_edu;
    spytech = (double)spy->spy_tech;
    spygoof = (double)errormargin(spy)/100.0;

    if (loctype==EF_SHIP) {
	ship = (struct shpstr *)loc;
	milown = ship->shp_own;
	mileff = (double)ship->shp_effic/100.0;
	x = ship->shp_x;
	y = ship->shp_y;
    } else {
	sect = (struct sctstr *)loc;
	milown = sect->sct_own;
	mileff = (double)sect->sct_effic/100.0;
	x = sect->sct_x;
	y = sect->sct_y;
    }

    natp = getnatp(cnum);
    relat = getrel(natp, milown);

    if (!milown || milown==spy->spy_own || relat==ALLIED) {
#ifdef REPORTZERORISK
	if (riskmsg) 
	    pr(fmt("%s%szero risk\n", riskmsg, *riskmsg ? ": ": ""));
#endif
	return 0;
    }

    natp = getnatp(milown);
    mil = getvar(V_MILIT, loc, loctype);
    miledu =  (double)natp->nat_level[NAT_ELEV];
    miltech = (double)natp->nat_level[NAT_TLEV];

    detect  = spygoof * mileff/spyeff * (double)difficulty/100.0
	* (double)mil/((mil<20) ? 6.0 : ((mil<100) ? 4.0 : 2.0))
	    * (miltech+1.0)/(spytech+1.0);
    
    risk    = spygoof * mileff/spyeff * (double)difficulty/100.0 
	* (double)mil/((mil<20) ? 15.0 : ((mil<100) ? 10.0 : 5.0))
	    * (miledu+0.1)/(spyedu+0.1)
		* (relat==NEUTRAL) ? 0.15 : ((relat==HOSTILE) ? 0.55 : 1.0);

    capture = mileff/spyeff - mileff;

    if (risk > 1.0) {
	spy->spy_hap *= (80+roll(20))/100;
	note = "suicidal";
    } else if (risk > 0.80) {
	spy->spy_hap *= (90+roll(10))/100;
	note = "extreme";
    } else if (risk > 0.45) {
	spy->spy_hap *= (95+roll(5))/100;
	note = "high";
    } else if (risk > 0.10) {
	note = "normal";
    } else {
	spy->spy_hap *= (99+roll(3))/100;
	note = "low";
    }

    if (riskmsg) {
	pr(fmt("%s%s%s risk (%d%% screw up)\n",
	       riskmsg, *riskmsg ? ": ": "", note,
	       (int)(spygoof*100.0)));
    }
    
    dpr(fmt("cantdoit: %s",spyas(spy,LOS)));
    dpr(fmt("  spy:          owner=%-3d edu=%-6.2f tech=%-6.2f eff=%.3f",
	    cnum,spyedu,spytech,spyeff));
    dpr(fmt("  %3d military: owner=%-3d edu=%-6.2f tech=%-6.2f eff=%.3f",
	    mil,milown,miledu,miltech,mileff));
    dpr(fmt("%.2f degree of difficulty",(double)difficulty/100.0));
    dpr(fmt("%.2f chance of goof-up",spygoof));
    dpr(fmt("Risk factors: detection (%.3f), attack (%.3f), capture (%.3f)\n",
	    detect,risk,capture));

    if (!chance(detect)) return 0;

    wu(0, milown, 
       fmt("Clandestine activity detected %s",
	   locasc(loc,loctype,LPS,milown)));
    if (!chance(risk)) {
	milown = 0;
    } else {
	i = (int)((double)mil * mileff * (double)roll(20)/5.0);
	dam = min(i,35);
	if (dam < 5) dam = 0;
	dpr(fmt("  %d mil inflict %d damage (%d)",mil,dam,i));
	spy->spy_effic -= dam;
	note = dam ? ((spy->spy_effic<5) ? "killed" : "wounded") : "shot at";
	pr(fmt(">BANG!< %s %s %s (down %d to %d%%)\n",
	       spyas(spy,0),note,locas(loc,loctype,LPOS), dam,spy->spy_effic));
	wu(0, milown, 
	   fmt("Military %s %s %s",
	       locasc(loc,loctype,LPS,milown),
	       note, spyas(spy,0)));
	spy->spy_hap = spy->spy_hap * (16+roll(4))/20;
	if (spy->spy_effic >= 5) {
	    if (chance(capture)) {
		pr(fmt("%s captured by local military.\n",spyas(spy,0)));
		wu(0, milown,
		   fmt("Your troops have captured %s",
		       spyasc(spy,LO,milown)));
		nreport(milown,N_CAPTURE_SPY,cnum,1);
		spy->spy_x = x;
		spy->spy_y = y;
		spy->spy_own = milown;
		spy->spy_effic = 10;
		spy->spy_mobil = 0;
		spy->spy_hap = 0;
	    }
	} else {
	    nreport(milown,N_KILL_SPY,cnum,1);
	}
    }
    putspy(spy->spy_uid,spy);
    return (milown);
}

/* _____________________________________________________________________
**|                                                                     |
**| errormargin - return spy's margin of error (range 0 to 100)         |
**|                                                                     |
**| The base accuracy factor (as listed in 'lib/global/spy.c')          |
**| represents the percentage error in reports from a 100% efficient    |
**| spy.  All the spy's personal factors are considered here.           |
**|_____________________________________________________________________|
*/

int errormargin(spy)
struct spystr *spy;
{
    extern int etu_per_update;
    extern int spy_retire;
    int i,j,r20,r80,age;

    age = spy->spy_age * etu_per_update;
    r20 = spy_retire / 5; /* 20% of career */
    r80 = r20 * 4;	  /* 80% of career */

    /* every 2% eff drop adds 1 point to acc */
    j = (100 - spy->spy_effic)/2; 

    /* accuracy increases through first 20% of career */
    if (age < r20) j += (10 * (r20-age)/r20);

    /* accuracy degrades in last 20% of career */
    if (age > r80) j += (20 * (age-r80)/r20);

    /* happy spies don't make as many mistakes */
    j -= (spy->spy_hap / 4);

    /* handicap manic depressives */
    if (spy->spy_hap < 4) j += (4 - spy->spy_hap)*2;

    /* smart spies are also more accurate */
    j -= (spy->spy_edu / 2);

    /* reduce error rate if he's high tech and knows how to exploit it */
    j -= (spy->spy_edu * spy->spy_tech / 100);

    i = spchr[spy->spy_type].sp_acc + j;
    if (i<2) i = 2; /* can't be more than 98% perfect */
    if (i>100) i = 100;

    return i;
}

placeown(loc,loctype)
char *loc;
int loctype;
{
    return (loctype==EF_SHIP) 
	? ((struct shpstr *)loc)->shp_own
	    : ((struct sctstr *)loc)->sct_own;
}

/*
**  putplace - do a `putsect' or a `putship' as necessary
**  (this doesn't really belong in this module)
*/

putplace(loc,loctype)
char *loc;
int loctype;
{
    struct shpstr *ship;
    struct sctstr *sect;
    if (loctype==EF_SHIP) {
	ship = (struct shpstr *)loc;
	putship(ship->shp_uid,ship);
    } else {
	sect = (struct sctstr *)loc;
	putsect(sect);
    }
}


