#ifndef lint
static char *RCSid = "$Header: navi.c,v 1.1 89/12/14 08:20:15 jay Exp $";
#endif

/*
 * navi.c
 *
 * navigate ships and such.
 *
 * from PSL Empire, 1985
 */

#include <ctype.h>
#include "misc.h"
#include "var.h"
#include "ship.h"
#include "sect.h"
#include "news.h"
#include "xy.h"
#include "nsc.h"
#include "nat.h"
#include "path.h"
#include "deity.h"
#include "file.h"

static int check_nav();
 /* return codes from check_nav */
#define CN_LANDLOCKED	(1)
#define CN_CONSTRUCTION	(2)
#define CN_ERROR	(-1)
#define CN_NAVIGABLE	(0)

#define sectis(x)	(sect.sct_type==(x))

char *bestpath();
double basemobcost();
extern double sub_slowdown;

nav_ship(sp,announcearrival)
register struct shpstr *sp;
int announcearrival;
{
    extern double techfact();
    extern char *argp[];
    struct sctstr sect;
    struct natstr *np;
    register m,n;
    char *cp,*adj;
    int	max,dir,nships;
    coord newx,newy,dx,dy;
    int stopping,snum;
    int mines,shells,mil;
    int	petcost,pet;
    double mobcost,mobil,petburnodo;
    struct nstr_item nb;
    struct mchrstr *mcp;
    int	vec[I_MAX+1];
    int	time_to_stop;
    int	owner,cnum;
    int nuclear;

    if (!sp->shp_own) return 0;

    /* just return if no autonaving to do for this ship */
    if (!(sp->shp_autonav & AN_AUTONAV) || (sp->shp_autonav & AN_STANDBY))
	return 0;

    cnum = sp->shp_own;
    snum = sp->shp_uid;
    mcp = &mchr[sp->shp_type];
    np = getnatp(cnum);
    mobil = (double) sp->shp_mobil;

    /* to be nuclear, the ship name must start with `nuc', it must
       have m_mpg=0, and must be able to carry rads */

    if (!strncmp(mcp->m_name,"nuc",3) && !mcp->m_mpg) 
	nuclear = vl_find(V_RAD,mcp->m_vtype,mcp->m_vamt,(int)mcp->m_nv);
    else
	nuclear = 0;
    pet = getvar(nuclear ? V_RAD : V_PETROL, (char *)sp, EF_SHIP);

    if (sp->shp_aboard > -1) {
	wu(0,cnum,
	   fmt("%s is being transported, put on standby\n",shipas(sp,0)));
	sp->shp_autonav |= AN_STANDBY;
	putship(sp->shp_uid, (char *)sp);
	return 0;
    }

    if (isballoon(sp) && sp->shp_depth==0) {
	wu(0,cnum,fmt("%s is on the ground, put on standby\n",shipas(sp,0)));
	sp->shp_autonav |= AN_STANDBY;
	putship(sp->shp_uid, (char *)sp);
	return 0;
    }

    if (mobil <= 0.0) {
	wu(0,cnum,fmt("%s has no mobility\n",shipas(sp,0)));
	return 0;
    }

    if (nuclear || mcp->m_mpg) {
	if (pet <= 0) {
	    wu(0,cnum,fmt("%s is out of fuel\n",shipas(sp,0)));
	    return 0;
	}
    }

    /* check crew - uws don't count */
    getvec(VT_ITEM, vec, (char *)sp, EF_SHIP);
    if (vec[I_MILIT] == 0 && vec[I_CIVIL] == 0) {
	wu(0,cnum,fmt("%s has no crew\n",shipas(sp,0)));
	return 0;
    }

    if (!getsect(sp->shp_x, sp->shp_y, &sect)) {
	wu(0,cnum,fmt("%s bad sector read\n",shipas(sp,0)));
	return 0;
    }
    
    owner = (sect.sct_own==sp->shp_own) ? 1 : 0;

    if (isship(sp)) {
	switch (check_nav(&sect)) {
	case CN_CONSTRUCTION:
	    wu(0, cnum, fmt("%s sector under construction\n",shipas(sp,0)));
	    return 0;
	case CN_LANDLOCKED:
	    wu(0, cnum, fmt("%s landlocked\n", shipas(sp,0)));
	    return 0;
	case CN_NAVIGABLE:
	    break;
	default:
	    wu(0, cnum, fmt("%s bad navigable\n", shipas(sp,0)));
	    return 0;
	}
    }
    
    stopping = 0;
    time_to_stop = 0;

    /* need to somehow make sector database available to bestpath */
    openbigmap(sp->shp_own);
    
    /* do an initial radar just in case we don't know what's around us */
    radmapnopr(sp->shp_x, sp->shp_y, (int)sp->shp_effic, 
	       (int)techfact(sp->shp_tech,
			     (double) mchr[sp->shp_type].m_vrnge),
	       (double)((mchr[sp->shp_type].m_flags & M_SONAR)
			? techfact(sp->shp_tech, 1.0) : 0.0));
    
    /* now try to calculate a route to the destination */
    cp = bestpath(sp->shp_x, sp->shp_y, sp->shp_destx,
		  sp->shp_desty,mcp->m_terrain);
    
    if ((*cp == '\0') || (*cp == '?')) {
	wu(0, cnum, fmt("%s bad path, ship put on standby\n",shipas(sp,0)));
	sp->shp_autonav |= AN_STANDBY;
	putship(sp->shp_uid, (char *)sp);
	return 0;
    }

    while (!stopping) {
	dx = dy = 0;
	dir = chkdir(*cp++, DIR_STOP, DIR_VIEW);
	if (dir == -1) 
	    break;
	else if (dir == DIR_STOP) 
	    stopping++;
	else {
	    dx = diroff[dir][0];
	    dy = diroff[dir][1];
	}
	newx = xnorm(sp->shp_x + dx);
	newy = ynorm(sp->shp_y + dy);
	getsect(newx, newy, &sect);
	if (stopping)
	    mobcost = 0.0;
	else {
	    
	    mobcost = basemobcost(sp,pet,nuclear);
	    
	    petcost = (int)dchr[sect.sct_type].d_mcst;
	    if (petcost == 0) petcost = 1;
	    if ((mcp->m_flags & M_ATV) && (petcost > 5)) petcost = 5;
	    mobcost *= petcost;
	    
	    petcost = 0;
	    if (!nuclear && mcp->m_mpg) {
		if (sp->shp_odom < 0) sp->shp_odom = 0;
		petburnodo = (double)(mchr[sp->shp_type].m_mpg)/10.0;
		petburnodo *= (double)(sp->shp_tech+200)/200.0;
		petburnodo *= (double)(sp->shp_effic)/100.0;
		if (petburnodo < 1.0) petburnodo = 1.0;
		if (petburnodo > 127.0) petburnodo = 127.0;
		petcost = ((sp->shp_odom+1) >= (int)petburnodo)?1:0;
	    }
	}
	
	owner = (sect.sct_own == sp->shp_own) ? 1 : 0;
	m = 0;
	
	/*
	**  If the next sector is impassable, issue a
	**  warning telegram and bail out now.
	*/

	adj = "";
	if (isship(sp)) {
	    if ( check_nav(&sect) != CN_NAVIGABLE ||
		(sect.sct_own && !owner && 
		 !sectis(SCT_BSPAN) && !sectis(SCT_TRADE) &&
		 getrel(getnatp(sect.sct_own),sp->shp_own)!=ALLIED)) {
		adj = "navigable";
	    } else {
		if (sp->shp_depth>0 && 
		    (sectis(SCT_TRADE) || sectis(SCT_HARBR))) {
		    wu(0,cnum,fmt("%s surfacing to enter %s\n",
				  shipas(sp,0),
				  xyas(sect.sct_x,sect.sct_y,cnum)));
		    sp->shp_depth = 0;
		    mobil--;
		}
	    }
	} else {
	    if (!validterrain(dchr[sect.sct_type].d_mnem,
			      mchr[sp->shp_type].m_terrain)) {
		adj = "passable";
	    }
	}
	if (*adj) {
	    if (!*cp || *cp=='h') { /*** 6/8/90 bf ***/
		wu(0,cnum,
		   fmt("%s destination not %s, put on standby\n",
		       shipas(sp,0), adj));
		sp->shp_autonav |= AN_STANDBY;
		putship(sp->shp_uid, (char *)sp);
	    }
	    else wu(0,cnum,fmt("%s sector not %s\n", shipas(sp,0), adj));
	    cp = 0;
	    break;
	}
	
	if (mobcost > 0.0) {
	    if (cannotleave(sp,&mobcost,&petcost,1,1)) {
		cp = 0;
		mobil -= mobcost;
		sp->shp_mobil = (int) mobil;
		sp->shp_autonav |= AN_STANDBY;
		if (!nuclear && mcp->m_mpg) {
		    pet -= petcost;
		    if (pet <= 0) {
			pet = 0;
			wu(0,cnum,
			   fmt("%s out of fuel at %s\n",
			       shipas(sp,0),xyas(newx,newy,cnum)));
		    }
		    putvar(V_PETROL,pet,(char *)sp, EF_SHIP);
		}
		wu(0,cnum,
		   fmt("%s put on standby\n",shipas(sp,0)));
		putship(sp->shp_uid, (char *)sp);
		break;
	    }
	}

	/*
	** JNP  12/15/89
	**   Stuff added for bigmap, generates automatic radar
	**   while moving 
	*/
	radmapnopr(newx, newy, (int)sp->shp_effic, 
		   (int)techfact(sp->shp_tech,
				 (double) mchr[sp->shp_type].m_vrnge),
		   (double)((mchr[sp->shp_type].m_flags & M_SONAR)
			    ? techfact(sp->shp_tech, 1.0) : 0.0));
	sp->shp_x = newx;
	sp->shp_y = newy;
	mobil -= mobcost;
	pet -= petcost;
	
	if (!nuclear && mcp->m_mpg)
	    sp->shp_odom = petcost ? 0 : sp->shp_odom+1;
	
	if ((mobil <= 0.0) || stopping || 
	    (!nuclear && mcp->m_mpg && (pet <= 0))) {
	    sp->shp_mobil = (int) mobil;
	    if (!nuclear && mcp->m_mpg) {
		if (pet <= 0) {
		    pet = 0;
		    sp->shp_autonav |= AN_STANDBY;
		    wu(0, cnum, 
		       fmt("%s out of fuel at %s, put on standby\n",
			   shipas(sp,0),xyas(newx,newy,cnum)));
		}
		putvar(V_PETROL,pet,(char *)sp, EF_SHIP);
	    }
	    putship(snum, sp);
	    time_to_stop = 1;
	    cp = &dirch[DIR_STOP];
	}

	if (stopping || !isship(sp) || sp->shp_depth>1) continue;
	
	owner = (sect.sct_own == sp->shp_own);
	mines = getvar(V_MINE, (char *)&sect, EF_SECTOR);

	if ((mcp->m_flags & M_SWEEP) && mines > 0 && !owner) {
	    max = vl_find(V_SHELL, mcp->m_vtype,
			  mcp->m_vamt, (int) mcp->m_nv);
	    shells = getvar(V_SHELL, (char *)sp, EF_SHIP);
	    for (m=0; mines > 0 && m < 5; m++) {
		if (chance(0.66)) {
		    mines--;
		    shells = min(max, shells + 1);
		    setbigmap(newx,newy,0x81);
		}
	    }
	    putvar(V_MINE, mines, (char *)&sect, EF_SECTOR);
	    putvar(V_SHELL, shells, (char *)sp, EF_SHIP);
	    putsect(&sect);
	}
	if (mines > 0 && !owner && 
	    chance(((mcp->m_flags & M_SWEEP) ? 0.03 : 0.05) * mines)) {
	    /* send a telegram instead of pr'ing */
	    wu(0, cnum, fmt("Mine detected by %s in %s!\n",
			    shipas(sp,0), xyas(newx, newy, cnum)));
	    nreport(cnum, N_HIT_MINE, 0, 1);
	    if (mcp->m_glim)
		m = 15 + 1350 / mcp->m_armor;
	    else
		m = 2100 / (mcp->m_vrnge*mcp->m_visib);
	    m += random() % 21;
	    shipdamage(sp, m);
	    mines--;
	    setbigmap(newx,newy,0x81);
	    putvar(V_MINE, mines, (char *)&sect, EF_SECTOR);
	    putsect(&sect);
	    cp = 0;
	    if (sp->shp_effic < 20) {
		time_to_stop = 1;
	    }
	    putship(sp->shp_uid, (char *)sp);
	    wu(0, cnum, fmt("%d%% damage sustained\n", m));
	}
	if (time_to_stop)
	    stopping = 1;
    }
    writebigmap();
    closebigmap();

    if (!sp->shp_own) {
	wu(0, cnum, fmt("%s owned by deity\n", shipas(sp,0)));
	return 1;
    }
    
    mcp = &mchr[sp->shp_type];
    
    if (mobil < 0) mobil = 0;
    sp->shp_mobil = mobil;
    
    if (pet < 0) pet = 0;
    if (!nuclear && mcp->m_mpg)
	putvar(V_PETROL,pet,(char *)sp,EF_SHIP);
    
    if ((newx == sp->shp_destx) && (newy == sp->shp_desty)) {
	sp->shp_autonav &= ~AN_AUTONAV;
	if (announcearrival)
	    wu(0, cnum, fmt("%s arrived at %s\n", 
			    shipas(sp,0),xyas(newx, newy, cnum)));
    }
    
    putship(snum, sp);
    return 1;
}

static int check_nav(sect)
struct sctstr *sect;
{
    extern struct dchrstr dchr[];
    switch (dchr[sect->sct_type].d_flg & 03) {
    case NAV_60: if (sect->sct_effic < 60) return CN_CONSTRUCTION;
    case NAV_02: if (sect->sct_effic < 2) return CN_CONSTRUCTION;
    case NAVOK:  break;
    default:  	 return CN_LANDLOCKED;
    }
    return CN_NAVIGABLE;
}


int
chkdir(dir_char, min_dir, max_dir)
        char    dir_char;
        int     min_dir;
        int     max_dir;
{
        register int i;

        for (i = min_dir; i <= max_dir; i++)
                if (dir_char == dirch[i])
                        return i;
        return -1;
}


double
techfact(level, mult)
        int     level;
        double  mult;
{
        return mult * ((50.0 + level) / (200.0 + level));
}


radmapnopr(cx, cy, eff, range, seesub)
	int	cx;
	int	cy;
	int	eff;
	int	range;
	double	seesub;
{
	int	rng;
	struct	sctstr sect;
	struct	shpstr ship;
	struct	nstr_sect ns;
	struct	nstr_item ni;
	int	x, y;
	int	row;
	int	n;

	range = (int) (range * (eff / 100.0));
	/* JNP 12/15/89  radar always can see at least 1 distant */
	if (range < 1) range = 1;
	snxtsct_dist(&ns, cx, cy, range);
	while (nxtsct(&ns, &sect)) {
		if (sect.sct_own == cnum ||
		    (sect.sct_type <= SCT_RURAL && sect.sct_type != SCT_SANCT)
		    || ns.curdist <= 1 /* 'looks' at adjacent sectors */
		    || ns.curdist <= range/3)
			setbigmap(ns.x, ns.y, dchr[sect.sct_type].d_mnem);
		else
			setbigmap(ns.x, ns.y, '?');
	}
	writebigmap();
}


snxtsct_dist(np, cx, cy, dist)
	register struct nstr_sect *np;
	coord	cx, cy;
	int	dist;
{
	bzero(np, sizeof(*np));
	xydist_range(cx, cy, dist, &np->range);
	np->cx = cx;
	np->cy = cy;
	np->ncond = 0;
	np->dist = dist;
	np->type = NS_DIST;
	np->read = ef_read;
	np->x = np->range.lx - 1;
	np->y = np->range.ly;
	np->dx = -1;
	np->dy = 0;
	xysize_range(&np->range);
	ef_zapcache(EF_SECTOR);
}

xysize_range(rp)
	register struct range *rp;
{
	if (rp->lx >= rp->hx)
		rp->width = WORLD_X + rp->hx - rp->lx;
	else
		rp->width = rp->hx - rp->lx;
	if (rp->ly >= rp->hy)
		rp->height = WORLD_Y + rp->hy - rp->ly;
	else
		rp->height = rp->hy - rp->ly;
}

xydist_range(x, y, dist, rp)
	coord	x;
	coord	y;
	int	dist;
	struct	range *rp;
{
	rp->lx = xnorm((coord)(x - 2 * dist));
	rp->hx = xnorm((coord)(x + 2 * dist) + 1);
	rp->ly = ynorm((coord)(y - dist));
	rp->hy = ynorm((coord)(y + dist) + 1);
	rp->height = 0;
	rp->width = 0;
}


int
nxtsct(np, sp)
	register struct nstr_sect *np;
	struct	sctstr *sp;
{
	while (1) {
		np->dx++;
		np->x++;
		if (np->x >= WORLD_X)
			np->x = 0;
		if (np->dx >= np->range.width) {
			np->dx = 0;
			np->x = np->range.lx;
			np->dy++;
			if (np->dy >= np->range.height)
				return 0;
			np->y++;
			if (np->y >= WORLD_Y)
				np->y = 0;
		}
		if ((np->y + np->x) & 01)
			continue;
		if (np->type == NS_DIST) {
			np->curdist = mapdist(np->x, np->y, np->cx, np->cy);
			if (np->curdist > np->dist)
				continue;
		}
		np->id = sctoff(np->x, np->y);
		if (!np->read(EF_SECTOR, np->id, (caddr_t) sp))
			continue;
		return 1;
	}
	/*NOTREACHED*/
}


double basemobcost(sp,rads,nuclear)
struct shpstr *sp;
int rads, nuclear;
{
    double mobcost;
    mobcost = (nuclear>0) ? ((double)rads/(double)nuclear) : 1.0;
    if (mobcost > 1.0) mobcost = 1.0;
    mobcost *= (sp->shp_effic * 0.01 * mchr[sp->shp_type].m_speed);
    mobcost = 480.0 /
	(mobcost + techfact(sp->shp_tech, mobcost));
    if ((mchr[sp->shp_type].m_flags & M_SUB) && sp->shp_depth>0) 
	mobcost *= sub_slowdown;
    return (mobcost);
}
