static char *RCSid = "$Header: land.c,v 1.1 89/12/14 13:31:33 jay Exp $";

/*
 * land.c
 *
 * new version of empcre
 * Create the land masses in the game.
 *
 * James W. Anderson, 1987
 *
 */

#include <stdio.h>
#include <fcntl.h>
#include "var.h"
#include "misc.h"
#include "power.h"
#include "nat.h"
#include "sect.h"
#include "gamesdef.h"
#include "file.h"
#include "xy.h"

struct powstr powbuf;

#define XSIZE		(WORLD_X / 2)
#define YSIZE		WORLD_Y
#define BIGV		256
#define SMALLV		128
#define TINYV		32

#define OCEAN	  0
#define ISLAND	  1
#define CONTINENT 2

#define dfltHIGHMIN 90
#define dfltPLATMIN 35
#define dfltHILLMIN 32
#define dfltLANDMIN 1
#define dfltPLAND   35
#define dfltPISLE   20
#define dfltCSIZE   3
#define dfltMETEORS 10
#define dfltMETSIZE 8

#define rnd(x) 		(random() % (x))
#define iabs(x)		(((x)<0) ? -(x) : (x))
#define xnorm(i)	((((i)<0) ? ((i)+XSIZE) : (i)) % XSIZE)
#define ynorm(i)	((((i)<0) ? ((i)+YSIZE) : (i)) % YSIZE)

int HIGHMIN,PLATMIN,HILLMIN,LANDMIN;
int NUMLAND,NUMISLE,NUMWATER,METEORS,METSIZE;
int percentland,percentisle,interactive;
int csize;
char  buf[80];

FILE   *sectf;
struct sctstr sects[YSIZE][XSIZE];

int    world[YSIZE][XSIZE];
int    plates[YSIZE][XSIZE];

configure(s)
char *s;
{	    
    int n;
    int ISLECH;
    char *t;
    if (!s) {
	HIGHMIN = dfltHIGHMIN;
	PLATMIN = dfltPLATMIN;
	HILLMIN = dfltHILLMIN;
	LANDMIN = dfltLANDMIN;
	METEORS = dfltMETEORS;
	METSIZE = dfltMETSIZE;
	csize   = dfltCSIZE;
	percentland  = dfltPLAND;
	percentisle  = dfltPISLE;
	interactive = 0;
    }
    for (t=s; *t && *t!='=' && *t!=':' && *t!=' '; t++);
    if (*t) t++;
    n = atoi(t);
    if (!strncmp(s,"el3",3)) HIGHMIN = n; else
    if (!strncmp(s,"el2",3)) PLATMIN = n; else
    if (!strncmp(s,"el1",3)) HILLMIN = n; else
    if (!strncmp(s,"el0",3)) LANDMIN = n; else
    if (!strncmp(s,"met",3)) METEORS = n; else
    if (!strncmp(s,"msz",3)) METSIZE = n; else    
    if (!strncmp(s,"lan",3)) percentland = n; else
    if (!strncmp(s,"isl",3)) percentisle = n; else
    if (!strncmp(s,"con",3)) csize = n; else
    if (s) {
	printf("You can choose from the following options:\n");
	printf("    el3:value   mountain elevation (0-127)\n");
	printf("    el2:value   plateau elevation (0-127)\n");
	printf("    el1:value   hill elevation (0-127)\n");
	printf("    el0:value   land elevation (0-127)\n");
	printf("    lan:value   percent of world that is land (0-100)\n");
	printf("    isl:value   island/land size ratio (percentage, 0-100)\n");
	printf("    con:value   continental seed size (typically 3)\n");
	printf("    met:value   number of meteor strikes (0-127)\n");
	printf("    msz:value   max meteor size (1-10)\n");
	interactive++;
    }
    NUMLAND  = (XSIZE*YSIZE*percentland)/100;
    NUMISLE  = NUMLAND*percentisle/100;
    NUMWATER = (XSIZE*YSIZE)-(NUMLAND+NUMISLE);
    printf("\n%s:\n",s ? "Current settings" : "INITIAL SETTINGS");
    printf("el3  mountain elev =   %d\n",HIGHMIN);
    printf("el2  plateau elev =    %d\n",PLATMIN);
    printf("el1  hill elev =       %d\n",HILLMIN);
    printf("el0  land elev =       %d\n",LANDMIN);
    printf("lan  land sectors =    %d (%d%%)\n",NUMLAND,percentland);
    printf("isl  island sectors =  %d (%d%%)\n",NUMISLE,percentisle);
    printf("con  continent seed =  %d\n",csize);
    printf("met  meteors strikes=  %d\n",METEORS);
    printf("msz  meteor size =     %d\n",METSIZE);
    printf("     water sectors =   %d\n",NUMWATER);
    printf("\n");
}
	
    
main(argc,argv)
int argc;
char *argv[];
{
    extern struct empfile empfile[];
    extern int maxnoc;
    int i,j,y1,x1,ybase,xbase;
    int elev[12+12+3]; /* # sects from -12 to 12 in steps of 10 elev */
    int sum,total;
    int range,rangesq;
    int sectrow,sectcol;
    int ifail,jfail,pass;
    long now;
    register int *irow;
    int fout;
    struct sctstr *sct;
    char *sectfil,*s;

    configure(0);
    for (j=1;j<argc && !interactive;j++) 
	configure(argv[j]);

    while (interactive) {
	*buf = 0;
	printf("option: "); 
	fflush(stdout);
	fgets(buf,80,stdin);
	if (*buf>' ') configure(buf);
	else interactive=0;
    }

    sectfil = empfile[EF_SECTOR].file;
    fout = open(sectfil, O_RDWR|O_CREAT|O_TRUNC, 0660);
    if (fout < 0) {
	perror(sectfil);
	exit(1);
    }

    time(&now);
    srandom(now+getpid());

    /* ________________________________________________________________
    **|                                                                |
    **| Flood the planet                                               |
    **|________________________________________________________________|
    */

    for (i=0; i < YSIZE; i++)
	for (j=0; j < XSIZE; j++)
	    plates[i][j] = OCEAN;

    /* ________________________________________________________________
    **|                                                                |
    **| Make continents                                                |
    **|________________________________________________________________|
    */

    for (pass=total=0; total<NUMLAND; ) {

retry1:
	if ((pass++ % 12) == 7) {

	    /* Place this one anywhere (may merge with another continent) */
	    while (plates[(i=rnd(YSIZE))][(j=rnd(XSIZE))] == OCEAN);

	} else {

	    i = rnd(YSIZE);
	    j = rnd(XSIZE);
	    /* Insure that continent seeds are at least 6 sectors apart */
	    for (ybase = i-6; ybase <= i+6; ybase++)
		for (xbase = j-6; xbase <= j+6; xbase++) {
		    x1 = xnorm(xbase);
		    y1 = ynorm(ybase);
		    if (plates[y1][x1] != OCEAN) goto retry1;
		}

	}

	for (ybase = i-csize; ybase <= i+csize; ybase++) {
	    ifail = iabs(i-ybase);
	    for (xbase = j-csize; xbase <= j+csize; xbase++) {
		x1 = xnorm(xbase);
		y1 = ynorm(ybase);
		if (plates[y1][x1] == OCEAN) {
		    jfail = iabs(j-x1);
		    if (jfail < ifail) jfail = ifail;
		    if (jfail<(csize-1) || !rnd(jfail+1) || !rnd(jfail+1)) {
			plates[y1][x1] = CONTINENT;
			printf("Continent at %d,%d\n",x1,y1);
			total++;
			break;
		    }
		}
	    }
	}
    }

    /* ________________________________________________________________
    **|                                                                |
    **| Seed islands                                                   |
    **|________________________________________________________________|
    */

    for (total=0; total<NUMISLE; total++) {
	while (plates[i=rnd(YSIZE)][j=rnd(XSIZE)] != OCEAN);
	plates[i][j] = ISLAND;
    }

    /* ________________________________________________________________
    **|                                                                |
    **| Now grow world from the seeded sectors                         |
    **|________________________________________________________________|
    */

    for (sectrow=0; sectrow < YSIZE; sectrow++) {
	ybase = sectrow*1;
	for (sectcol=0; sectcol < XSIZE; sectcol++) {
	    xbase = sectcol*1;

	    total = plates[sectrow][sectcol];
	    for (i = ybase+1; i-- > ybase;) {
		for (j = xbase+1; j-- > xbase; )
		    switch (total) {
		    case OCEAN:     world[i][j] = rnd(SMALLV)-rnd(BIGV);   break;
		    case ISLAND:    world[i][j] = rnd(BIGV) - rnd(BIGV)+2; break;
		    case CONTINENT: world[i][j] = rnd(BIGV) - rnd(SMALLV); break;
		    }
	    }
	}
    }

    /* ________________________________________________________________
    **|                                                                |
    **| Lob in a few meteoric catastrophes                             |
    **|________________________________________________________________|
    */

    for (pass=0; pass < METEORS; pass++) {

	i = rnd(YSIZE);
	j = rnd(XSIZE);
	range = 3+rnd(METSIZE);
	range *= range;

	for (ybase = i-range; ybase<=i+range; ybase++) {
	    for (xbase = j-range; xbase<=j+range; xbase++) {
		x1 = xbase-j; x1 *= x1;
		y1 = ybase-i; y1 *= y1;
		rangesq = x1+y1;
		if (rangesq < range) {
		    j = xnorm(xbase);
		    i = ynorm(ybase);
		    sum = 100 * (range-rangesq)/range;
		    sum = rnd(TINYV)+sum;
		    world[i][j] -= sum;
		}
	    }
	}
    }

    /* ________________________________________________________________
    **|                                                                |
    **| Assign elevations, mountains, etc.                             |
    **|________________________________________________________________|
    */

    sum = 0;
    sct = (struct sctstr *) sects;

    for (i=0; i < 12+12+3; i++)
 	elev[i] = 0;

    for (i=0; i<YSIZE; i++) {
	for (j=0; j<XSIZE; j++, sct++) {

	    sct->sct_x = j*2 + (i & 01);
	    sct->sct_y = i;
	    range = 3 + rnd(2);
	    rangesq = range * range;
	    total = 0;
	    for (ybase = i + range; ybase-- > i;) {
		irow = world[ybase % YSIZE];
		for (xbase = j + range; xbase-- > j;)
		    total += irow[xbase % XSIZE];
	    }

	    total /= rangesq;
	    if (total < LANDMIN) 	sct->sct_type = SCT_WATER;
	    else if (total < HILLMIN)	sct->sct_type = SCT_RURAL;
	    else if (total < PLATMIN)	sct->sct_type = SCT_MOUNT;
	    else if (total < HIGHMIN)	sct->sct_type = SCT_RURAL;
	    else			sct->sct_type = SCT_MOUNT;

	    sct->sct_elev = total;
	    sct->sct_newtype = sct->sct_type;
	    sct->sct_harden = 0;
	    sum += total;

	    if (total < -129)     elev[0]++;
	    else if (total > 129) elev[26]++;
	    else                  elev[13+total/10]++;
	}
    }

    /* ________________________________________________________________
    **|                                                                |
    **| Write final file                                               |
    **|________________________________________________________________|
    */

    write(fout, sects, sizeof(sects));
    close(fout);
    for (i = 0; i < 12+12+3; i++)
	if (elev[i] != 0)
	    printf("%4d sectors elevation %4d to %4d\n",
		   elev[i], 10*i - 140, 10*i - 130);
    exit(0);
}
